from .n0struct_n0dict__ import n0dict__
from .n0struct_logging import n0pretty

# **********************************************************************************************************************
# *****************************************************************************
# https://www.w3.org/TR/html4/sgml/entities.html#h-24.4.1
# https://www.htmlhelp.com/reference/html40/entities/special.html
html_entities = {
    # C0 Controls and Basic Latin
    0x22:           "&quot;",   # quotation mark = APL quote
    0x26:           "&amp;",    # ampersand
    0x3C:           "&lt;",     # less-than sign
    0x3E:           "&gt;",     # greater-than sign
    # Latin Extended-A
    0x152:          "&OElig;",  # latin capital ligature OE
    0x153:          "&oelig;",  # latin small ligature oe
    # ligature is a misnomer, this is a separate character in some languages
    0x160:          "&Scaron;", # latin capital letter S with caron
    0x161:          "&scaron;", # latin small letter s with caron
    0x178:          "&Yuml;",   # latin capital letter Y with diaeresis
    # Spacing Modifier Letters
    0x02C6:         "&circ;",   # modifier letter circumflex accent
    0x02DC:         "&tilde;",  # small tilde, U+02DC ISOdia
    # General Punctuation
    0x2002:         "&ensp;",   # en space
    0x2003:         "&emsp;",   # em space
    0x2009:         "&thinsp;", # thin space
    0x200C:         "&zwnj;",   # zero width non-joiner
    0x200D:         "&zwj;",    # zero width joiner
    0x200E:         "&lrm;",    # left-to-right mark
    0x200F:         "&rlm;",    # right-to-left mark
    0x2013:         "&ndash;",  # en dash
    0x2014:         "&mdash;",  # em dash
    0x2018:         "&lsquo;",  # left single quotation mark
    0x2019:         "&rsquo;",  # right single quotation mark
    0x201A:         "&sbquo;",  # single low-9 quotation mark
    0x201C:         "&ldquo;",  # left double quotation mark
    0x201D:         "&rdquo;",  # right double quotation mark
    0x201E:         "&bdquo;",  # double low-9 quotation mark
    0x2020:         "&dagger;", # dagger
    0x2021:         "&Dagger;", # double dagger
    0x2030:         "&permil;", # per mille sign
    0x2039:         "&lsaquo;", # single left-pointing angle quotation mark, it is proposed but not yet ISO standardized
    0x203A:         "&rsaquo;", # single right-pointing angle quotation mark, it is proposed but not yet ISO standardized
    0x20AC:         "&euro;",   # euro sign, U+20AC NEW
}
# **********************************************************************************************************************
class n0dict_(n0dict__):
    # *************************************************************************
    # XPATH
    # *************************************************************************
    def __xpath(self, value, path: str = None, mode: int = None) -> list:
        """
        Private function: recursively collect elements xpath starts from parent
        """
        result = []
        if isinstance(value, (list, tuple)):
            for i, subitm in enumerate(value):
                result += self.__xpath(subitm, "%s[%d]" % (path, i), mode)
        elif isinstance(value, dict):
            for key, value in value.items():
                result += self.__xpath(value, "%s/%s" % (path, key), mode)
        elif isinstance(value, (str, int, float)) or value is None:
            result.append((path, value))
        else:
            raise Exception("Not expected type (%s) %s/%s == %s" % (type(value), path, key, str(value)))
        return result
    # **************************************************************************
    def xpath(self, mode: int = None) -> list:  # list[(xpath, value)]
        """
        Public function: collect elements xpath starts from root
        """
        return self.__xpath(self, "/", mode)
    # **************************************************************************
    def to_xpath(self, mode: int = None) -> str:
        """
        Public function: collect elements xpath starts from root and print with indents
        """
        result = ""
        xpath_list = self.xpath(mode)
        if xpath_list:
            xpath_maxlen = max(len(itm[0]) for itm in xpath_list) + 2  # plus 2 chars '"]'
            for itm in xpath_list:
                result += ("['%-" + str(xpath_maxlen) + "s = %s\n") % \
                            (
                                itm[0] + "']",  # Don't move to the main
                                ('"' + str(itm[1]) + '"') if itm[1] else "None"
                            )
        return result
    # **************************************************************************
    # XML
    # **************************************************************************
    def __xml(self, parent: dict, indent: int, inc_indent: int) -> str:
        """
        Private function: recursively export n0dict into xml result string
        """
        result = ""
        if not parent is None:
            if isinstance(parent, dict):
                if not len(parent.items()):
                    return ""
                for key, value in parent.items():
                    if result:
                        result += "\n"
                    if isinstance(value, (list, tuple)):
                        for i, subitm in enumerate(value):
                            if i:
                                result += "\n"
                            sub_result = self.__xml(subitm, indent + inc_indent, inc_indent)
                            if not sub_result:
                                if isinstance(sub_result, str):
                                    result += " " * indent + "<%s></%s>" % (key,key)
                                else:
                                    result += " " * indent + "<%s/>" % key
                            else:
                                if '>' in sub_result:
                                    result += (" " * indent + "<%s>\n%s\n" + " " * indent + "</%s>") % (key, sub_result, key)
                                else:
                                    result += (" " * indent + "<%s>%s</%s>") % (key, sub_result, key)
                    elif isinstance(value, (str, int, float)):
                        if not key.startswith("@"):
                            result += " " * indent + ("<%s>%s</%s>" % (key, str(value).translate(html_entities), key))
                    elif isinstance(value, dict):
                        sub_result = self.__xml(value, indent + inc_indent, inc_indent)

                        attribs = ""
                        attribs_of_current_key = [(__key[1:], __value) for __key,__value in value.items() if __key.startswith("@")]
                        if len(attribs_of_current_key):
                            for __key, __value in attribs_of_current_key:
                                attribs += " %s=\"%s\"" % (__key, __value)
                        if sub_result:
                            result += (" " * indent + "<%s%s>\n%s\n" + " " * indent + "</%s>") % (key, attribs, sub_result, key)
                        else:
                            result += " " * indent + "<%s%s/>" % (key, attribs)
                    elif value is None:
                        result += " " * indent + "<%s/>" % key
                    else:
                        raise Exception("__xml(..): Unknown type (%s) %s ==  %s" % (type(value), key, str(value)))
            elif isinstance(parent, (list, tuple)):
                if not len(parent):
                    return None
                for i, itm in enumerate(parent):
                    if i:
                        result += "\n"
                    result += self.__xml(itm, indent + inc_indent, inc_indent)
            elif isinstance(parent, (str, int, float)):
                result += str(parent)
            else:
                print("Exception")
                raise Exception("__xml(..): Unknown type (%s) ==  %s" % (type(parent), str(parent)))

            return result
        else:
            # return None
            return ""
    # **************************************************************************
    def to_xml(self, indent: int = 4, encoding: str = "utf-8") -> str:
        """
        Public function: export self into xml result string
        """
        result = ""
        if encoding:
            result = "<?xml version=\"1.0\" encoding=\"%s\"?>\n" % encoding
        return result + self.__xml(self, 0, indent)
    # **************************************************************************
    # JSON
    # **************************************************************************
    def __json(self, parent: dict, indent: int, inc_indent: int) -> str:
        """
        Private function: recursively export n0dict into json result string
        """
        result = ""
        for key, value in parent.items():
            if result:
                result += ",\n"
            if isinstance(value, list):
                sub_result = ""
                for i, subitm in enumerate(value):
                    if sub_result:
                        sub_result += ",\n"
                    sub_sub_result = self.__json(subitm, indent + inc_indent * 2, inc_indent)
                    if sub_sub_result:
                        if isinstance(subitm, dict):
                            sub_result += (" " * (indent + inc_indent) + "{\n%s\n" + " " * (
                                    indent + inc_indent) + "}") % sub_sub_result
                        elif isinstance(subitm, (list, tuple)):
                            sub_result += (" " * (indent + inc_indent) + "[\n%s\n" + " " * (
                                    indent + inc_indent) + "]") % sub_sub_result
                if sub_result:
                    result += (" " * indent + '"%s": [\n%s\n' + " " * indent + "]") % (key, sub_result)
                else:
                    result += " " * indent + '"%s": null' % key
            elif isinstance(value, str):
                result += " " * indent + ('"%s": "%s"' % (key, value))
            elif isinstance(value, dict):
                sub_result = self.__json(value, indent + inc_indent, inc_indent)
                if sub_result:
                    result += (" " * indent + '"%s": {\n%s\n' + " " * indent + "}") % (key, sub_result)
                else:
                    result +=  " " * indent + '"%s": null' % key
            elif value is None:
                result += " " * indent + '"%s": null' % key
            else:
                raise Exception("Unknown type (%s) %s ==  %s" % (type(value), key, str(value)))
        return result
    # **************************************************************************
    def to_json(self, 
                indent: int = 4,
                pairs_in_one_line = True,
                json_convention: bool = True,
                skip_empty_arrays: bool = False,
                compress: bool = False,
    ) -> str:
        """
        Public function: export self into json result string
        """
        if compress:
            indent = 0
        
        return n0pretty(self,
                        show_type=False,
                        auto_quotes=False,
                        __quotes='"',
                        __indent_size=indent,
                        pairs_in_one_line=pairs_in_one_line,
                        json_convention=json_convention,
                        skip_empty_arrays=skip_empty_arrays,
                        show_item_count=False,
        )
# ******************************************************************************
# ******************************************************************************



import typing
# import urllib # expected_value = urllib.parse.unquote(expected_value) # mypy: error: Module has no attribute "parse"
from urllib.parse import unquote as urllib__parse__unquote
# ******************************************************************************
# ******************************************************************************
def split_name_index(node_name: str) -> typing.Tuple[
                                                        # typing.Any,
                                                        str,
                                                        typing.Union[
                                                            str,
                                                            typing.Tuple[str, str, typing.Union[str, bool]]
                                                        ]
                                                    ]:
    if not isinstance(node_name, str):
        raise Exception("node_name (%s)%s must be string" % (type(node_name), node_name))

    node_index_tuple = None
    if '[' in node_name and node_name.endswith(']'):
        node_name, node_index_str = node_name[:-1].split('[', 1)
        node_name = node_name.strip()
        node_index_str = node_index_str.strip()
        if isinstance(node_index_str, str):
            if node_index_str != "":
                if node_index_str.lower().startswith('contains') and node_index_str.endswith(')'):
                    node_index_part1, node_index_part2 = node_index_str[8:-1].strip().split('(',1)[1].split(',',1)
                    if node_index_part1.lower().startswith('text'):
                        node_index_str = "text()~~" + node_index_part2
                if '=' in node_index_str or '~' in node_index_str:
                    separators = ("==","!=","~~","!~","~","=")
                    for separator in separators:
                        if separator in node_index_str:
                            expected_node_name, expected_value = node_index_str.split(separator,1)
                            expected_node_name = expected_node_name.strip()
                            expected_value = expected_value.strip()
                            if separator == '=':
                                separator = '=='
                            if separator == '~':
                                separator = '~~'
                            break
                    else:
                        raise Exception("Never must be happend!")

                    expected_value_bool = False  # Default value, in real None was used, but mypy raised error
                    if expected_value.lower() == "true()":
                        expected_value = ""
                        expected_value_bool = True
                    elif expected_value.lower() == "false()":
                        expected_value = ""
                        expected_value_bool = False
                    elif (expected_value.startswith('"') and expected_value.endswith('"')) or \
                            (expected_value.startswith("'") and expected_value.endswith("'")):
                        expected_value = expected_value[1:-1]
                        # expected_value = urllib.parse.unquote(expected_value) # mypy: error: Module has no attribute "parse"
                        expected_value = urllib__parse__unquote(expected_value)
                    node_index_tuple = (expected_node_name, separator, expected_value or expected_value_bool)
    else:
        node_index_str = None
    return node_name, (node_index_tuple if not node_index_tuple is None else node_index_str)
# ******************************************************************************
# notemptyitems(item):
#   Check item or recursively subitems of item.
#   Return count of notempty item/subitems.
# ******************************************************************************
def notemptyitems(item):
    not_empty_items_count = 0
    # if isinstance(item, (dict, OrderedDict, n0dict)):
    if isinstance(item, dict):
        for key in item:
            not_empty_items_count += notemptyitems(item[key])
    elif isinstance(item, (list, tuple)):
        for itm in item:
            not_empty_items_count += notemptyitems(itm)
    else:
        if item:
            not_empty_items_count += 1
    return not_empty_items_count
# ******************************************************************************
# ******************************************************************************



import typing
from .n0struct_utils import n0eval
from .n0struct_findall import findall as n0struct_findall__findall
from .n0struct_findall import findfirst as n0struct_findall__findfirst
# ******************************************************************************
# ******************************************************************************
class n0list_(list):
    def findall(self, xpath: str, raise_exception: bool = True):
        return n0struct_findall__findall(self, xpath, raise_exception)
    def findfirst(self, xpath: str, raise_exception: bool = True):
        return n0struct_findall__findfirst(self, xpath, raise_exception)
    # **************************************************************************
    # **************************************************************************
    # n0list_. _get()
    # **************************************************************************
    def _get(self, xpath: typing.Union[str, int], raise_exception = True, if_not_found = None, return_lists = True):
        """
        Private function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, exception IndexError will be raised
        """
        if isinstance(xpath, int):
            if (xpath < 0 and -xpath <= len(self)) or (xpath >= 0 and xpath < len(self)):
                return self[xpath]
            else:
                return if_not_found
        elif not isinstance(xpath, str) or not xpath:
            return if_not_found

        if xpath.startswith('?'):
            xpath = xpath[1:]
            raise_exception = False
            if_not_found = ''
        if any(char in xpath for char in "/["):
            parent_node, node_name_index, cur_value, xpath_found_str, not_found_xpath_list = self._find(xpath, self, return_lists)
            if not not_found_xpath_list:
                return cur_value
            else:
                if raise_exception:
                    raise IndexError("not found '%s' in '%s'" % ('/'.join(not_found_xpath_list), xpath_found_str))
                else:
                    return if_not_found
        else:
            try:
                return super(n0list_, self).__getitem__(n0eval(xpath))
            except IndexError as ex:
                if raise_exception:
                    raise ex
                else:
                    return if_not_found
    # **************************************************************************
    # n0list_. get()
    # **************************************************************************
    def get(self, xpath: typing.Union[str, int], if_not_found = None):
        """
        Public function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, if_not_found will be returned
        """
        return self._get(xpath, raise_exception = False, if_not_found = if_not_found)
    # **************************************************************************
    # n0list_. first()
    # **************************************************************************
    def first(self, xpath: str, if_not_found = None):
        """
        Public function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, if_not_found will be returned
        If self[where1/where2/.../whereN] is list, thet the first element will be returned
        """
        result = self._get(xpath, raise_exception = False, if_not_found = if_not_found, return_lists = False)
        if isinstance(result, (list, tuple)) and len(result) == 1:
            result = result[0]
        return result
    # **************************************************************************
    def __getitem__(self, xpath):
        """
        Public function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, exception IndexError will be raised
        """
        if isinstance(xpath, str):
            return self._get(xpath, raise_exception = True)
        else:
            return super(n0list_, self).__getitem__(xpath)
    # **************************************************************************
    # def append(self, sigle_item):
    # if isinstance(sigle_item, (list,n0list_)):
    # raise (TypeError, '(%s)%s must be scalar' % (type(sigle_item), sigle_item))
    # super(n0list_, self).append(sigle_item)  #append the item to itself (the list)
    # return self
    # **************************************************************************
    # def extend(self, other_list):
    # if not isinstance(other_list, (list,n0list_)):
    # raise (TypeError, '(%s)%s must be list' % (type(sigle_item), sigle_item))
    # super(n0list_, self).extend(other_list)
    # return self
    # **************************************************************************
    def _in(self, other_list, in_is_expected: bool):
        if not isinstance(other_list, (list, tuple)):
            other_list = [other_list]
        for itm in self:
            if (itm in other_list) == in_is_expected:
                return True
        else:
            return False
    # **************************************************************************
    def any_in(self, other_list):
        return self._in(other_list, True)
    # **************************************************************************
    def any_not_in(self, other_list):
        return not self._in(other_list, True)
    # **************************************************************************
    def all_in(self, other_list):
        return not self._in(other_list, False)
    # **************************************************************************
    def all_not_in(self, other_list):
        return self._in(other_list, False)
    # **************************************************************************
    def _consists_of(self, other_list, in_is_expected: bool):
        if not isinstance(other_list, (list, tuple)):
            other_list = [other_list]
        for itm in other_list:
            if super(n0list_, self).__contains__(itm) == in_is_expected:
                return True
        else:
            return False
    # **************************************************************************
    def consists_of_any(self, other_list):
        return self._consists_of(other_list, True)
    # **************************************************************************
    def consists_of_all(self, other_list):
        return not self._consists_of(other_list, False)
    # **************************************************************************
    def __contains__(self, other_list):  # otherlist in n0list_([a])
        return self.consists_of_all(other_list)
    # **************************************************************************
    def not_consists_of_any(self, other_list):
        return not self._consists_of(other_list, True)
# ******************************************************************************
# ******************************************************************************



import typing
# ******************************************************************************
# ******************************************************************************
__flag_compare_check_different_types = False
def set__flag_compare_check_different_types(value: bool):
    """
    if __flag_compare_check_different_types == True, then
    validate type of attributes in .compare()/.direct_compare()
    and return result["difftypes"]
    """
    global __flag_compare_check_different_types
    __flag_compare_check_different_types = value
    return __flag_compare_check_different_types
def get__flag_compare_check_different_types() -> bool:
    global __flag_compare_check_different_types
    return __flag_compare_check_different_types
# ******************************************************************************
__flag_compare_return_difference_of_values = False
def set__flag_compare_return_difference_of_values(value: bool):
    """
    if __flag_compare_return_difference_of_values == True, then
    if values of attributes are different and are int,float,
    return additional element in result["not_equal"] with difference
    """
    global __flag_compare_return_difference_of_values
    __flag_compare_return_difference_of_values = value
def get__flag_compare_return_difference_of_values() -> bool:
    global __flag_compare_return_difference_of_values
    return __flag_compare_return_difference_of_values
# ******************************************************************************
__flag_compare_return_equal = False
def set__flag_compare_return_equal(value: bool):
    """
    if __flag_compare_return_equal == True, then
    if records are the same, 
    then return additional element in result["equal"] with equal records
    """
    global __flag_compare_return_equal
    __flag_compare_return_equal = value
def get__flag_compare_return_equal() -> bool:
    global __flag_compare_return_equal
    return __flag_compare_return_equal
# ******************************************************************************
__flag_compare_return_place = True
def set__flag_compare_return_place(value: bool):
    """
    if __flag_compare_return_place == True, then
        if record is the different, then return tupple of index and record in result["self_unique"]/result["other_unique"]
    else
        return only the record in result["self_unique"]/result["other_unique"]
    """
    global __flag_compare_return_place
    __flag_compare_return_place = value
def get__flag_compare_return_place() -> bool:
    global __flag_compare_return_place
    return __flag_compare_return_place
# ******************************************************************************
# xpath_match(xpath: str, xpath_list):
#   Check that real xpath (or xpath like) is equal any of xpath_list[0..n].
#   Returns i+1
# ******************************************************************************
def xpath_match(xpath: str, xpath_list: typing.Union[str, list, tuple]) -> int:
    """
    :param xpath:
        xpath: str
    :param xpath_list:
        xpath_list: str or list|tuple
    :return:
         0 not matched any of xpath_list
        >0 matched
    """
    if isinstance(xpath_list, str):
        # xpath_list = (xpath_list,)
        xpath_list = [xpath_list]  # 0.18 = 2020-10-22 workaround fix for Py38: changing (A,) into [A], because of generates NOT tuple, but initial A
    if not isinstance(xpath_list, (tuple, list)):
        raise Exception("xpath_match(..): unknown type of xpath_list = %s" % type(xpath_list))

    xpath_parts = xpath.split("/")
    for i, xpath_itm in enumerate(xpath_list):
        xpath_itm_parts = xpath_itm.split("/")
        for j, part in enumerate(reversed(xpath_itm_parts)):
            if not part:  # //
                # n0print("MATCH: matched relative")
                return i + 1
            # n0debug("part")
            if j >= len(xpath_parts):
                # n0print("not matched: too short")
                break
            if part != "*" and part.lower() != xpath_parts[-1 - j].lower():  # /*/
                # n0print("not matched: not equal")
                break
        else:
            # n0print("MATCH: matched full")
            return i + 1
        # n0print("Let's try new loop")
    # n0print("not matched: not matched with all from list")
    return 0
# ******************************************************************************
def generate_composite_keys(
                            input_list: list,
                            elements_for_composite_key: tuple,
                            prefix: str = None,
                            transform: typing.List[typing.Tuple[str, typing.Callable[[str], str]]] = []
                            ) -> list:
    """
    serialization all or {elements_for_composite_key} elements of {input_list[]}
    :param transform: ()|None|empty mean nothing to transform, else [[<xpath to elem>,<lambda for transformatio>],..]
    :return:
        [[<composite_key>,[<index of entry>],...}
    """
    if isinstance(elements_for_composite_key, str):
        elements_for_composite_key = [elements_for_composite_key]  # 0.18 = 2020-10-22 workaround fix for Py38: changing (A,) into [A], because of generates NOT tuple, but initial A

    composite_keys_for_all_lines = []
    if transform:
        attributes_to_transform = [itm[0] for itm in transform]
    else:
        attributes_to_transform = []

    for line_i, line in enumerate(input_list):
        # if isinstance(line, (dict, OrderedDict, n0dict)):
        if isinstance(line, dict):
            created_composite_key = ""
            if elements_for_composite_key:
                for key in elements_for_composite_key:
                    if key in line:
                        if created_composite_key:
                            created_composite_key += ";"
                        fullxpath = "%s/%s" % (prefix, key)
                        transform_i = xpath_match(fullxpath, attributes_to_transform)
                        if transform_i:
                            transform_i -= 1
                            tranformed = transform[transform_i][1](line[key])
                        else:
                            tranformed = str(line[key])
                        created_composite_key += key + "=" + tranformed
        else:
            raise Exception("generate_composite_keys(..): expected element dict inside list, but got (%s)%s" % (type(line), line))
        composite_keys_for_all_lines.append((created_composite_key, line_i))
    return composite_keys_for_all_lines
# ******************************************************************************
# ******************************************************************************



import typing
from .n0struct_utils import n0isnumeric
from .n0struct_logging import *
# ******************************************************************************
# ******************************************************************************
def findall(current_node: typing.Union[dict, list], seeked_xpath_str: str, raise_exception = True) -> dict:
    if not isinstance(seeked_xpath_str, str):
        raise TypeError(f"seeked_xpath_str: expected str, got other type: {type(seeked_xpath_str)}{str(seeked_xpath_str)}")
    if seeked_xpath_str.startswith("./"):
        seeked_xpath_str = seeked_xpath_str[2:]
    if seeked_xpath_str.startswith("//"):
        seeked_xpath_str = seeked_xpath_str[2:]
    if seeked_xpath_str.startswith("/"):
        seeked_xpath_str = seeked_xpath_str[1:]
    seeked_xpath_list = [
                        stripped_item
                        for itm in seeked_xpath_str \
                                                    .replace("[", '/[') \
                                                    .replace("//", "/") \
                                                    .split('/')
                        if (stripped_item:=itm.strip())
    ]
    return _findall(current_node, seeked_xpath_list)
# ******************************************************************************
def findfirst(current_node: typing.Union[dict, list], seeked_xpath_str: str, raise_exception = True) -> tuple:
    found = findall(current_node, seeked_xpath_str, False)
    if not found:
        if raise_exception:
            raise IndexError(f"Not found item {seeked_xpath_str}")
        else:
            return None,None
    elif len(found) != 1:
        if raise_exception:
            raise IndexError(f"Found more that single item {seeked_xpath_str}")
            
    found_keys = list(found.keys())
    return found_keys[0], found[found_keys[0]]
# ******************************************************************************
def _findall(
            parent_node: typing.Union[dict, list],
            seeked_xpath_list: list,

            found_xpath_list: list = [],
            parent_nodes_stack: dict = {}, # {found_xpath_list[0]: parent_node^0..found_xpath_list[0..-1]: parent_node^-1}

            raise_exception = True,
            level = 0,
) -> dict:
    """
        returns:
            {
                seeked_xpath_str: node_ptr # if node_ptr is dict/list it's real node, else the final element
            }
    """
    # # n0print("="*80)
    # # n0debug("level")
    # # n0print("="*80)
    # # # n0debug("found_xpath_list")
    # # n0debug_calc("//" + "/".join(found_xpath_list).replace('/[', '['), "found_xpath_list")
    # # n0debug("parent_node")
    # # # n0debug("seeked_xpath_list")
    # # n0debug_calc("/".join(seeked_xpath_list).replace('/[', '['), "seeked_xpath_list")
    # # n0debug("parent_nodes_stack")
    # # n0print("-"*80)

    if not seeked_xpath_list:
        # # n0print("*"*30 + f" FOUND!!! Force surface...")
        # # n0debug_calc("//" + "/".join(found_xpath_list).replace('/[', '['), "found_xpath_list")
        # # n0debug("parent_node")
        return {"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}
    # **************************************************************************
    # *** Arguments validation
    ##if not isinstance(parent_node, (dict, list)):
    ##    if raise_exception:
    ##        raise TypeError(f"parent_node: expected dict/list, got other type: {type(parent_node)} '{str(parent_node)}'")
    ##    else:
    ##        return None
    if not isinstance(seeked_xpath_list, list):
        if raise_exception:
            raise TypeError(f"seeked_xpath_list: expected list, got other type: {type(seeked_xpath_list)} '{str(seeked_xpath_list)}'")
        else:
            return None
    if not isinstance(found_xpath_list, list):
        if raise_exception:
            raise TypeError(f"found_xpath_list: expected list, got other type: {type(found_xpath_list)} '{str(found_xpath_list)}'")
        else:
            return None
    # **************************************************************************
    # Get child name or index/condition
    # **************************************************************************
    child_name = None
    child_index = None
    if seeked_xpath_list[0].strip() == '..':
        # # n0print("*"*30 + f" Surface...")
        if len(parent_nodes_stack) < 2:
            if raise_exception:
                raise KeyError(f"Imposible to surface from {found_xpath_list+'/'+child_name}")
            else:
                return None
        '''
        # **********************************************************************
        # Surface
        if child_name == "..":
            # ******************************************************************
            # Surface for one level
            n0debug_calc(parent_nodes_stack,         "parent_nodes_stack")
            n0debug_calc(parent_nodes_stack[-1],     "parent_nodes_stack[-1]")

            n0debug_calc(seeked_xpath_list,             "seeked_xpath_list")
            n0debug_calc(seeked_xpath_list[1:],         "seeked_xpath_list[1:]")

            n0debug_calc(found_xpath_list,       "found_xpath_list")
            n0debug_calc(found_xpath_list[-1],   "found_xpath_list[-1]")

            n0debug_calc(parent_nodes_stack,         "parent_nodes_stack")
            n0debug_calc(parent_nodes_stack[:-1],    "parent_nodes_stack[:-1]")

            n0debug_calc(raise_exception,        "raise_exception")

            return _findall(
                            parent_nodes_stack[-1],
                            seeked_xpath_list[1:],
                            found_xpath_list[-1],
                            parent_nodes_stack[:-1],
                            raise_exception,
                            level + 1,
            )
        '''
        # n0debug("parent_nodes_stack")
        del parent_nodes_stack[list(parent_nodes_stack.keys())[-1]]
        # n0debug("parent_nodes_stack")
        # n0debug("found_xpath_list")
        # # n0debug_calc("//" + "/".join(found_xpath_list[:-1]).replace('/[', '['), "found_xpath_list[:-1]")
        # # n0debug_calc(parent_nodes_stack[list(parent_nodes_stack.keys())[-1]], "parent_nodes_stack[list(parent_nodes_stack.keys())[-1]]")
        # # n0debug_calc("/".join(seeked_xpath_list[1:]).replace('/[', '['), "seeked_xpath_list[1:]")
        return _findall(
                        parent_nodes_stack[list(parent_nodes_stack.keys())[-1]],
                        # parent_nodes_stack,
                        seeked_xpath_list[1:],
                        found_xpath_list[:-1],
                        # {**parent_nodes_stack, **{"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}},
                        parent_nodes_stack,
                        raise_exception,
                        level - 1,
        )
    elif seeked_xpath_list[0].startswith("["):
        # **********************************************************************
        # Index or condition
        if not seeked_xpath_list[0].endswith("]"):
            raise TypeError(f"Index or condition should be inside [], but '{str(seeked_xpath_list[0])}'")
        child_index = seeked_xpath_list[0][1:-1].strip() # Index or xpath condition/function
        if n0isnumeric(child_index):
            # Index
            child_index = int(child_index)
        else:
            # Conditions
            lower_child_index = child_index.lower().replace(' ','')
            # # n0debug("lower_child_index")
            if lower_child_index == '*':
                pass  # will be processed lately
            elif lower_child_index.startswith("last()"):
                after_last = lower_child_index[6:]
                child_index = eval("-1"
                                   + after_last if not any(not ch in "-+01234567890" for ch in after_last) else ""
                )
            elif lower_child_index.startswith("text()"):
                after_text = lower_child_index[6:]
                equal_condition = None
                if after_text.startswith('=='):
                    equal_condition = True
                    condition_separator = '=='
                elif after_text.startswith('='):
                    equal_condition = True
                    condition_separator = '='
                elif after_text.startswith('!='):
                    equal_condition = False
                    condition_separator = '!='
                elif after_text.startswith('<>'):
                    equal_condition = False
                    condition_separator = '<>'
                else:
                    raise TypeError(f"Unknown logic condition [{child_index}] in '{str(seeked_xpath_list[0])}'")
                before_condition,after_condition = child_index.split(condition_separator, 1)
                value_for_condition = after_condition.strip()
                if len(value_for_condition) > 1 and \
                   (
                       (value_for_condition.startswith('"') and value_for_condition.endswith('"'))
                       or
                       (value_for_condition.startswith("'") and value_for_condition.endswith("'"))
                   ):
                    value_for_condition = value_for_condition[1:-1]
                if (parent_node.lower() == value_for_condition.lower()) != equal_condition:
                    # # n0print(f"CONDITION FAILED: {child_index}: {parent_node} {'==' if equal_condition else '!='} {value_for_condition}")
                    return None
                else:
                    # # n0print("*"*30 + f" Deep after condition [{child_index}]...")
                    found_xpath_list[-1] += seeked_xpath_list[0]
                    return _findall(
                                    parent_node,
                                    seeked_xpath_list[1:],
                                    # found_xpath_list[-1] + seeked_xpath_list[0:1],
                                    found_xpath_list,
                                    {**parent_nodes_stack, **{"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}},
                                    raise_exception,
                                    level + 1,
                    )
            else:
                raise TypeError(f"Unknown condition [{child_index}] in '{str(seeked_xpath_list[0])}'")
    else:
        child_name = seeked_xpath_list[0]
        # # n0debug("child_name")
        # # n0debug("parent_node")
    # **************************************************************************
    if isinstance(parent_node, list):
        # **********************************************************************
        # If we have NOT got index for the list, so try to check all items in the list
        if child_index is None:
            # n0print("*"*30 + " NOT TESTED #0...")
            # n0debug("parent_node")
            # __seeked_xpath_list = "/".join(["[*]"] + seeked_xpath_list)
            # n0print("FOUND: " + "/".join(found_xpath_list))
            # n0print(f"========================= {level} -> {__seeked_xpath_list}")
            # result = _findall(
                            # parent_node,
                            # ["[*]"] + seeked_xpath_list,
                            # found_xpath_list,
                            # parent_nodes_stack,
                            # raise_exception,
                            # level,
            # )
            # n0print(f"-------------------------")
            # n0debug("result")
            # n0print(f"-------------------------")
            # exit()
            # return result
            return _findall(
                            parent_node,
                            ["[*]"] + seeked_xpath_list,
                            found_xpath_list,
                            parent_nodes_stack,
                            raise_exception,
                            level,
            )
        elif isinstance(child_index, int):
            if child_index >= len(parent_node) or (child_index < 0 and -child_index > len(parent_node)):
                if raise_exception:
                    raise IndexError(f"Index {child_index} is out of range {len(parent_node)} at '{'//' + '/'.join(found_xpath_list).replace('/[', '[')}'")
                else:
                    return None
            child_node = parent_node[child_index]
            if isinstance(child_node, (dict, list)):
                n0print("*"*30 + " NOT TESTED #1...")
                found_xpath_list[-1] += f"[{child_index}]"
                return _findall(
                                child_node,
                                seeked_xpath_list[1:],
                                # found_xpath_list + seeked_xpath_list[0:1],
                                # found_xpath_list + [f"[{child_index}]"],
                                found_xpath_list,
                                # parent_nodes_stack + [(found_xpath_list, parent_node)],
                                {**parent_nodes_stack, **{"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}},
                                raise_exception,
                                level + 1,
                )
            else:
                if raise_exception:
                    raise IndexError(f"Child under {'//' + '/'.join(found_xpath_list).replace('/[', '[')}[{child_index}] is {type(child_node)}'{child_node}', but expected dict/list")
                else:
                    return None
        elif isinstance(child_index, str) and child_index == '*':
            # # n0print("*"*30 + " Multiple deep to each index [*]...")
            multi_found = {}
            if not len(found_xpath_list):
                found_xpath_list = [""]
            last_xpath = found_xpath_list[-1]
            for child_index, child_node in enumerate(parent_node):
                if isinstance(child_node, (dict, list)):
                    # # n0print("*"*25 + f" Multiple deep to index [{child_index}]...")
                    found_xpath_list[-1] = last_xpath + f"[{child_index}]"
                    found = _findall(
                                    # child_node,
                                    __parent_node := parent_node[child_index],
                                    __seeked_xpth := seeked_xpath_list[1:],
                                    # found_xpath_list + [f"[{child_index}]"],
                                    __found_xpath := found_xpath_list,
                                    # parent_nodes_stack + [(found_xpath_list, parent_node)],
                                    __parent_stck := {**parent_nodes_stack, **{"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}},
                                    raise_exception,
                                    level,
                    )
                    # n0debug("__parent_node")
                    # n0print("seeked_xpath_list:" + "/".join(__seeked_xpth))
                    # n0print("found_xpath_list:" + "/".join(__found_xpath))
                    # n0debug("__parent_stck")
                    # n0debug("found")
                    if found:
                        multi_found.update(found)
                else:
                    if raise_exception:
                        raise IndexError(f"Child under {'//' + '/'.join(found_xpath_list).replace('/[', '[')}[{child_index}] is {type(child_node)}'{child_node}', but expected dict/list")
                    else:
                        return None
            return multi_found
        else:
            raise NotImplementedError(f"Not implemented xpath language: '{seeked_xpath_list[0]}'") # xpath parser will implemented in the next version
    # **************************************************************************
    # parent_node[child_name]
    elif isinstance(parent_node, dict):
        # **********************************************************************
        # Just to be sure that for dict we have name of subnode and no index
        if child_index:
            if raise_exception:
                raise IndexError(f"Index {child_index} could not be uplied to {type(parent_node)} at '{found_xpath_list}'")
            else:
                return None
        if not child_name:
            if raise_exception:
                raise KeyError(f"Subnode {child_node} is not found at '{found_xpath_list}'")
            else:
                return None
        # **********************************************************************
        # Find any
        if child_name == '*':
            multi_found = {}
            # ******************************************************************
            # just skip * and check if the current node is satisfied to the xpath/conditions
            ## n0print("*"*30 + " NOT TESTED #2...")
            found = _findall(
                            parent_node,
                            seeked_xpath_list[1:],
                            found_xpath_list,
                            parent_nodes_stack,
                            raise_exception,
                            level + 1,
            )
            if found:
                multi_found.update(found)
            # ******************************************************************
            # and just after check recursively all branches below
            for child_name in parent_node: # checking * sub-nodes
                child_node = parent_node[child_name]
                if isinstance(child_node, (dict, list)):
                    ## n0print("*"*30 + " NOT TESTED #3...")
                    # n0debug("child_node")
                    n0debug("seeked_xpath_list")
                    n0debug_calc(found_xpath_list + [child_name], "found_xpath_list + [child_name]") ## n0debug_calc(found_xpath_list + seeked_xpath_list[0:1], "found_xpath_list + seeked_xpath_list[0:1]")
                    n0debug_calc("/".join(found_xpath_list).replace('/[', '['), "parent_node")
                    found = _findall(
                                    child_node,
                                    seeked_xpath_list,
                                    # found_xpath_list + [child_name],
                                    found_xpath_list + [child_name], ##found_xpath_list + seeked_xpath_list[0:1],
                                    # parent_nodes_stack + [(found_xpath_list, parent_node)],
                                    {**parent_nodes_stack, **{"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}},
                                    raise_exception,
                                    level + 1,
                    )
                    if found:
                        multi_found.update(found)
            return multi_found
        # **********************************************************************
        # Find by exact xpath
        if child_name in parent_node:
            # # n0print("*"*30 + f" Deep to dict key [{child_name}]...")
            return _findall(
                            parent_node[child_name],
                            seeked_xpath_list[1:],
                            # found_xpath_list + [child_name],
                            found_xpath_list + seeked_xpath_list[0:1],
                            {**parent_nodes_stack, **{"//" + "/".join(found_xpath_list).replace('/[', '['): parent_node}},
                            raise_exception,
                            level + 1,
            )
        else:
            return None
    # **************************************************************************
    else:
        n0print("="*80)
        n0debug("level")
        n0print("="*80)
        # n0debug("found_xpath_list")
        n0debug_calc("//" + "/".join(found_xpath_list).replace('/[', '['), "found_xpath_list")
        n0debug("parent_node")
        # n0debug("seeked_xpath_list")
        n0debug_calc("/".join(seeked_xpath_list).replace('/[', '['), "seeked_xpath_list")
        # n0debug("parent_nodes_stack")
        n0print("-"*80)
        raise KeyError(f"Internal error:: looking for {seeked_xpath_list} already found: {found_xpath_list} in {type(parent_node)}'{str(parent_node)}'")
# ******************************************************************************
# ******************************************************************************



import os
import typing
# Used by class Git():
import subprocess
import signal
from .n0struct_logging import *
# ******************************************************************************
# ******************************************************************************
class Git():
    _repository_name = None
    _repository_path = None
    # ##############################################################################################
    def __init__(self, repo_root_dir: str, repository_url: str, rsa_key_path: str = ""):
        if not repository_url.startswith("ssh://") or not repository_url.endswith(".git"):
            raise Exception("repository_url must be 'ssh://...git'")

        outs = errs = None
        self._repository_path = os.path.abspath(repo_root_dir)
        if not os.path.exists(os.path.join(self._repository_path, ".git")):
            outs, errs = self.run(
                                ["clone", repository_url] +
                                ["--config", "core.sshCommand=ssh -i " + rsa_key_path + " -F /dev/null"] if rsa_key_path else []
            )
            self._repository_name = repository_url.split("/")[-1].split(".git")[0]
            self._repository_path = os.path.join(self._repository_path, self._repository_name)
        else:
            self._repository_name = os.path.split(self._repository_path)[1]
            n0print("Other repository '%s' is already existed" % self._repository_name)

        if errs and "already exists and is not an empty directory." in errs:
            outs, errs = self.run("pull")
            if outs != "Already up to date.\n":
                n0debug_calc(outs.strip(), "outs")
                n0debug_calc(errs.strip(), "errs")
                # raise Exception(outs + "\n" if outs else "" + errs)
    # ##############################################################################################
    def run(self, git_arguments: typing.Union[str, list], show_result = True) -> tuple:
        if isinstance(git_arguments, str):
            git_arguments = git_arguments.split(" ")
        n0print("*** git %s" % " ".join(git_arguments))
        # p = subprocess.Popen(   (command_line:=["git",] + git_arguments), # Only for 3.8+
        command_line = ["git",]
        p = subprocess.Popen(   (command_line + git_arguments),
                                cwd = self._repository_path,
                                stdout = subprocess.PIPE,
                                stderr = subprocess.PIPE,
                                shell = True,
                                encoding = "utf-8",
        )
        try:
            # outs, errs = p.communicate(timeout=(timeout_sec:=600)) # Only for 3.8+
            timeout_sec = 600
            outs, errs = p.communicate(timeout = timeout_sec)
        except subprocess.TimeoutExpired:
            raise Exception("Timeout %d seconds were happened during execution:\n%s>%s" % (timeout_sec, self._repository_path, " ".join(command_line)))
            os.kill(p.pid, signal.CTRL_BREAK_EVENT)
            outs, errs = p.communicate()

        if show_result:
            n0debug_calc(outs.strip(), "outs")
            n0debug_calc(errs.strip(), "errs")
        return outs, errs
    # ##############################################################################################
    def checkout(self, branch_name: str):
        outs, errs = self.run(["checkout", branch_name])
        return outs, errs
    # ##############################################################################################
    def log(self, git_arguments: typing.Union[str, list]):
        if isinstance(git_arguments, str):
            git_arguments = git_arguments.split(" ")

        outs, errs = self.run(["log",  "--date=format:%y%m%d_%H%M%S", "--pretty=format:%H=%ad=%cn=%s"] + git_arguments)
        return outs, errs
    # ##############################################################################################
# ******************************************************************************
# ******************************************************************************



from __future__ import annotations  # Python 3.7+: for using own class name inside body of class
import typing
from .n0struct_utils_find import split_name_index
from .n0struct_findall import findall as n0struct_findall__findall
from .n0struct_findall import findfirst as n0struct_findall__findfirst
from .n0struct_files import n0eval
from .n0struct_logging import *
# ******************************************************************************
# ******************************************************************************
class n0dict__(dict):
    def findall(self, xpath: str, raise_exception: bool = True):
        return n0struct_findall__findall(self, xpath, raise_exception)
    def findfirst(self, xpath: str, raise_exception: bool = True):
        return n0struct_findall__findfirst(self, xpath, raise_exception)
    # **************************************************************************
    # **************************************************************************
    def _get(self, xpath: str, raise_exception = True, if_not_found = None, return_lists = True):
        """
        Private function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, exception IndexError will be raised
        """
        if not xpath:
            raise Exception("xpath '%s' is not valid" % str(xpath))
        if xpath.startswith('?'):
            xpath = xpath[1:]
            raise_exception = False
            if_not_found = ''
        if any(char in xpath for char in "/["):
            parent_node, node_name_index, cur_value, xpath_found_str, not_found_xpath_list = self._find(xpath, self, return_lists)
            if not not_found_xpath_list:
                return cur_value
            else:
                if raise_exception:
                    raise IndexError("not found '%s' in '%s'" % ('/'.join(not_found_xpath_list), xpath_found_str))
                else:
                    return if_not_found
        else:
            try:
                return super(n0dict__, self).__getitem__(xpath)
            except KeyError as ex:
                if raise_exception:
                    raise ex
                else:
                    return if_not_found
    # **************************************************************************
    # **************************************************************************
    def get(self, xpath: str, if_not_found = None):
        """
        Public function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, if_not_found will be returned
        """
        return self._get(xpath, raise_exception = False, if_not_found = if_not_found)
    # **************************************************************************
    # **************************************************************************
    def first(self, xpath: str, if_not_found = None):
        """
        Public function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, if_not_found will be returned
        If self[where1/where2/.../whereN] is list, thet the first element will be returned
        """
        result = self._get(xpath, raise_exception = False, if_not_found = if_not_found, return_lists = False)
        if isinstance(result, (list, tuple)) and len(result) == 1:
            result = result[0]
        return result
    # ******************************************************************************
    # ******************************************************************************
    def __getitem__(self, xpath):
        """
        Public function:
        return self[where1/where2/.../whereN]
            AKA
        return self[where1][where2]...[whereN]

        If any of [where1][where2]...[whereN] are not found, exception IndexError will be raised
        """
        return self._get(xpath, raise_exception = True)
    # **************************************************************************
    # **************************************************************************
    def __setitem__(self, xpath: str, new_value):
        """
        Public function:
        self[where1/where2/.../whereN] = value
            AKA
        self[where1][where2]...[whereN] = value

        if xpath will be started with ?, the nothing will be done if new_value is None or empty
        if xpath is exist, then the value will be overwritten
        if not exist, then the node will be created

        could be used predicates:
            [<item>=<value>]
            [<item>='<value>']
        or indexes
            [0]
            [1]
            [2]
            [-1]
            [-2]
            [-3]
        or functions
            [last()]
            [last()-1]
            [last()-2]
        or commands for creating (convertion into list) new node
            [new()]
        """
        if xpath.startswith('?'):
            if new_value is None or (isinstance(new_value, str) and new_value == ""):
                return None
            xpath = xpath[1:]

        if any(char in xpath for char in "/["):
            parent_node, node_name_index, cur_value, xpath_found_str, not_found_xpath_list = self._find(xpath, self, return_lists = True)
            if not_found_xpath_list:
                parent_node, node_name_index = self._add(parent_node, node_name_index, not_found_xpath_list)

            node_name, node_index = split_name_index(node_name_index)
            if isinstance(parent_node, dict):
                if node_index:
                    raise Exception("How is it possible: index '%s' for dictionary (%s)'%s'?"% (node_index, type(parent_node), parent_node))
                parent_node.update({node_name_index: new_value})
            elif isinstance(parent_node, (list, tuple)):
                if node_name:
                    raise Exception("How is it possible: key '%s' for list (%s)'%s'?" % (node_name, type(parent_node), parent_node))
                eval_node_index = n0eval(node_index)
                parent_node[n0eval(node_index)] = new_value
            else:
                raise Exception("How is it possible: unknown type of parent node (%s) of '%s'" % (type(parent_node), parent_node))
        else:
            super(n0dict__, self).__setitem__(xpath, new_value)

        return new_value  # For speed
    # **************************************************************************
    def delete(self, xpath: str, recursively: bool = False) -> n0dict__:
        xpath_list = xpath.split('/')
        for i,last_xpath_index in enumerate(range(len(xpath_list), 0, -1)):
            parent_node, node_name_index, cur_value, xpath_found_str, not_found_xpath_list = \
                self._find(xpath_list[0:last_xpath_index], self, return_lists=True)
            if i == 0 or (
                recursively and
                isinstance(cur_value, n0dict__) and not len(cur_value)
            ):
                # n0debug("parent_node")
                # n0debug("node_name_index")
                if isinstance(parent_node, list) and not isinstance(node_name_index, int):
                    if not isinstance(node_name_index, str) or not node_name_index.startswith('[') or not node_name_index.endswith(']'):
                        raise Exception(f"Not expactable index for list {node_name_index}")
                    node_name_index = n0eval(node_name_index[1:-1])
                    # n0debug("node_name_index")
                del parent_node[node_name_index]
        return self
    # **************************************************************************
    def pop(self, xpath: str, recursively: bool = False) -> typing.Any:
        result = self[xpath]
        self.delete(xpath, recursively)
        return result
    # **************************************************************************
    # **************************************************************************
    def _valid(self, validate, valid_is_expected: bool):
        for itm in self:
            if validate(itm) == valid_is_expected:
                return True
        return False
    # **************************************************************************
    def any_valid(self, validate):
        return self._consists_of(validate, True)
    # **************************************************************************
    def any_not_valid(self, validate):
        return self._consists_of(validate, False)
    # **************************************************************************
    def all_valid(self, validate):
        return not self._consists_of(validate, False)
    # **************************************************************************
    def all_not_valid(self, validate):
        return not self._consists_of(validate, True)
    # **************************************************************************
    def valid(self, node_xpath:str, validate, expected_result_for_error: bool = False, msg:str = None):
        """
        :param node_xpath:
            xpath to the node inside self
        :param validate:
            list/scalar/function = validation
        :param expected_result_for_error:
            By default expected that if result of validation is True, then self[node_xpath] is not valid (return False)
        :param msg:
            if None => return result as bool True(validation)/False
        :return:

        Examples:
            xml.valid('node/subnode', ["",None], True, "ERROR")
                If xml['node/subnode'] is equal "" or None (result of comparising is True), then return ERROR, else ""
            xml.valid('node/subnode', ["",None], True)
                If xml['node/subnode'] is equal "" or None (result of comparising is True), then return False (not valid), else True
            xml.valid('node/subnode', "", True)
                If xml['node/subnode'] is equal "" (result of comparising is True), then return False (not valid), else True
            xml.valid('node/subnode', [1,2], False, "ERROR")
                If xml['node/subnode'] is not equal 1 or 2 (result of comparising is False), then return ERROR, else ""
            xml.valid('node/subnode', [1,2], False)
                If xml['node/subnode'] is not equal 1 or 2 (result of comparising is False), then return False (not valid), else True
            xml.valid('node/subnode', [1,2])
                If xml['node/subnode'] is not equal 1 or 2 (result of comparising is False), then return False (not valid), else True
        """
        try:
            node_value = self.get(node_xpath)
            if callable(validate):
                validate_result = validate(node_value)
            elif isinstance(validate, (list, tuple)):
                validate_result = node_value in validate
            else:
                validate_result = node_value == validate
        except:
            validate_result = False

        if validate_result == expected_result_for_error:
            if msg is None:
                return False
            else:
                return msg.format(node_xpath, str(node_value))
        else:
            if msg is None:
                return True
            else:
                return ""
    # **************************************************************************
    # **************************************************************************
    def update_extend(self, other):
        if other is None:
            return self
        elif isinstance(other, dict):
            for key in other:
                if key not in self:
                    self.update({key: other[key]})
                else:
                    if not isinstance(self[key], list):
                        self[key] = list(self[key])
                    if isinstance(other[key], (list, tuple)):
                        self[key].extend(other[key])
                    else:
                        self[key].append(other[key])
        elif isinstance(other, (str, int, float)):
            key = list(self.items())[0][0]  # [0]= first item, [0] = key
            self[key].append(other)
        elif isinstance(other, (list, tuple)):
            key = list(self.items())[0][0]  # [0]= first item, [0] = key
            for itm in other:
                if isinstance(itm, (list, tuple)):
                    self[key].extend(itm)
                else:
                    self[key].append(itm)
        else:
            raise Exception("Unexpected type of other: " + str(type(other)))
        return self
    # **************************************************************************
    # **************************************************************************
    def isExist(self, xpath) -> dict:
        """
        Public function: return empty lists in dict, if self[xpath] exists
        """
        validation_results = dict({
            "differences":  [],
            "not_equal":    [],
            "self_unique":  [],
            "other_unique": [],
        })
        if get__flag_compare_check_different_types():
            validation_results.update({"difftypes": []})

        # TO DO: redo with 'in'
        try:
            if self[xpath]:
                return validation_results
        except:
            pass
        validation_results["differences"].append("[%s] doesn't exist" % xpath)
        validation_results["other_unique"].append((xpath, None))
        return validation_results
    # **************************************************************************
    # **************************************************************************
    def is_exist(self, xpath: str) -> bool:
        """
        Public function: return True, if self[xpath] exists
        """
        # TO DO: redo with 'in'
        '''
        try:
            if self[xpath]:
                return True
        except:
            pass
        return False
        '''
        with contextlib.suppress(Exception):
            if self[xpath]:
                return True
        return False
    # **************************************************************************
    # **************************************************************************
    def has_all(self,tupple_of_keys):
        for key in tupple_of_keys:
            if key not in self:
                return False
            else:
                if self[key] is None:
                    return False
                if isinstance(self[key],(str,tuple,list,set,frozenset,dict)) and len(self[key]) == 0:
                    return False
        return True
    # **************************************************************************
    # **************************************************************************
    def has_any_of(self,tupple_of_keys):
        for key in tupple_of_keys:
            if key in self:
                return True
        return True
    # **************************************************************************
    # **************************************************************************
    def isEqual(self, xpath, value):
        """
        Public function: return empty lists in dict, if self[xpath] == value
        """
        validation_results = self.isExist(xpath)
        if notemptyitems(validation_results):
            return validation_results
        try:
            if self[xpath] == value:
                return []
        except:
            pass
        validation_results["differences"].append("[%s]=='%s' != '%s'" % (xpath, self[xpath], value))
        validation_results["not_equal"].append((xpath, (self[xpath], value)))
        return validation_results
    # **************************************************************************
    # **************************************************************************
    def isTheSame(self, xpath, other_n0dict, other_xpath=None, transformation=lambda x: x):
        """
        Public function: return empty lists in dict, if transformation(self[xpath]) == transformation(other_n0dict[other_xpath])
        """
        if not other_xpath:
            other_xpath = xpath
        validation_results = self.isExist(xpath).update_extend(other_n0dict.isExist(other_xpath))
        if notemptyitems(validation_results):
            return validation_results
        try:
            if transformation(self[xpath]) == transformation(other_n0dict[other_xpath]):
                return validation_results
        except:
            # n0print("EXCEPTION in 'if transformation(self[xpath]) == transformation(other_n0dict[other_xpath]):'")
            pass
        validation_results["differences"].append("[%s]=='%s' != [%s]=='%s'" % (
            xpath, transformation(self[xpath]),
            other_xpath, transformation(other_n0dict[other_xpath])
        )
                                              )
        validation_results["not_equal"].append((xpath, (self[xpath], other_n0dict[other_xpath])))
        return validation_results
# ******************************************************************************
# ******************************************************************************



from collections.abc import MutableSet
# ******************************************************************************
# ******************************************************************************
# https://stackoverflow.com/questions/1653970/does-python-have-an-ordered-set
# https://code.activestate.com/recipes/576694/
class OrderedSet(MutableSet):
    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]         # sentinel node for doubly linked list
        self.map = {}                   # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)
# ******************************************************************************
# 2 levels dict:
# initial_dict:
#   key1:
#       - item1
#       - item2
#   key2:
#       - key1
#       - item3
#       - item4
# Unpack references recursive: bool = True
# unpacked_dict = {key : list(unpack_references(initial_dict, key)) for key in initial_dict}
# unpacked_dict:
#   key1:
#       - item1
#       - item2
#   key2:
#       - item1
#       - item2
#       - item3
#       - item4
# Remove references recursive: bool = False
# unpacked_dict = {key : list(unpack_references(initial_dict, key), False) for key in initial_dict}
# unpacked_dict:
#   key1:
#       - item1
#       - item2
#   key2:
#       - item3
#       - item4
# ******************************************************************************
def unpack_references(initial_dict: dict, initial_key: str, recursive: bool = True) -> OrderedSet:
    collected_set = OrderedSet() # Not allow to save order with ordinary set
    node = initial_dict[initial_key]
    if isinstance(node, str):
        node = [node]
    if not isinstance(node, list):
        raise Exception(f"node under {initial_key} must be str or list")

    for item in node:
        if item in initial_dict:    # item == reference (key)
            if recursive:
                collected_set |= unpack_references(initial_dict, item)
        else:                       # item == component dir
            collected_set.add(item)
    return collected_set
# ******************************************************************************
# ******************************************************************************



import typing
from pathlib import Path
import n0struct
# from .n0struct_logging import *
from .n0struct_utils import n0eval
# ******************************************************************************
# ******************************************************************************
def load_file(file_name: str) -> str:
    with open(file_name, 'rt') as in_file:
        return in_file.read()
# ******************************************************************************
def load_lines(file_name: str) -> list:
    with open(file_name, 'rt') as in_file:
        # return [line.strip() for line in in_file.read().split("\n") if line.strip()]
        while True:
            if not (line:=in_file.readline()):
                break
            yield line.rstrip('\n')
# ******************************************************************************
def save_file(file_name: str, lines: typing.Any):
    Path(file_name).parent.mkdir(parents=True, exist_ok=True)

    if isinstance(lines, (list, tuple)):
        buffer = "\n".join(lines)
    elif isinstance(lines, str):
        buffer = lines
    else:
        buffer = str(lines)
    with open(file_name, 'wt') as out_filehandler:
        out_filehandler.write(buffer)
# ******************************************************************************
def load_ini(file_name: str, default_value = None, equal_sign = '=') -> dict:
    ini_dict = {}
    for line in load_file(file_name):
        items = line.strip().split(equal_sign, 1)
        if items:
            ini_dict.update({
                            items[0].upper():   (
                                                int(items[1])
                                                if len(items) > 1 and items[1].isdecimal()
                                                else default_value
                                                )
            })
    return ini_dict
# ******************************************************************************
def load_serialized(file_name: str,
                    # /,  # When everybody migrates to py3.8, then we will make it much beautiful
                    equal_tag: str = "=",
                    separator_tag: str = ";",
                    comment_tags: typing.Union[tuple, list] = ("#", "//"),
                    remove_startswith: str = "",
                    remove_endswith: str = ""
                    # ) -> n0struct.n0list:
                    ) -> list:
    result = []
    for line in load_file(file_name):
        line = line.strip()
        if any(line.startswith(comment_tag) for comment_tag in comment_tags):
            continue
        if line.startswith(remove_startswith):
            line = line[len(remove_startswith):]
        if line.startswith(remove_startswith):
            line = line[len(remove_startswith):]

        pairs = line.split(separator_tag)
        if len(pairs):
            result.append({})
            for pair in pairs:
                if equal_tag in pair:
                    tag, value = pair.split(equal_tag, 1)
                    value = value.strip()
                    if (value.startswith('"') and value.endswith('"')) or \
                       (value.startswith("'") and value.endswith("'")):
                        value = value[1:-1]
                    else:
                        value = n0eval(value)
                else:
                    tag = pair
                    value = None
                result[-1].update({tag.strip(): value})
                # result["root"][-1].update({tag.strip(): value})
    return result
# ******************************************************************************
# ******************************************************************************



# -*- coding: utf-8 -*-
# 0.01 = 2020-07-25 = Initial version
# 0.02 = 2020-07-26 = Enhancements
# 0.03 = 2020-08-02 = Huge enhancements
# 0.04 = 2020-08-05 = Prepared for upload to pypi.org
# 0.05 = 2020-08-11 = Huge enhancements: unification of .*compare() .toJson(), .toXml(), n0dict(JSON/XML string)
# 0.06
# 0.07 = 2020-09-02 = .compare(transform=..) and .direct_compare(transform=..) added
# 0.08 = 2020-09-05 = refactoring
# 0.09 = lost
# 0.10 = rewritten date related functions
# 0.11 = 2020-09-17 fixed returning time by date_delta(..)
# 0.12 = 2020-09-24 added function to find keys in n0dict
# 0.13 = 2020-10-11
#        n0dict.nvl(..) now supports complicated path like A/B/C
#        n0dict.__FindElem(..) previously supports modificators [i], [last()], [last()-X], [-X]
#                              now additionally supports modificators [text()="XYZ"], [text()=="XYZ"], [text()!="XYZ"], /../
#                              XYZ must be encoded with urlencode
# 0.14 = 2020-10-17 n0print prints to stderr
# 0.15 = 2020-10-19 n0pretty(..): fix for json.decoder.JSONDecodeError: Expecting property name enclosed in double quotes
#                   strip_namespaces() is added
# 0.16 = 2020-10-20 strip_namespaces() transformed into transform_structure()
#                   n0dict.compare() is fixed to use compare_only=
# 0.17 = 2020-10-22 n0print(..): optimization
# 0.18 = 2020-10-22 workaround fix for Py38: changing (A,) into [A], because of generates NOT tuple, but initial A
# 0.19 = 2020-10-24 fixed issue with autotests, recursive convertion is added into constructor
# 0.20 = 2020-10-26 get_composite_keys(transform=..) is added, numeric checking is fixed
# 0.21 = 2020-11-09 fixed Exception: Why parent is None?
#                   date_slash_ddmmyyyy() is added
# 0.22 = 2020-11-09 fixed date_now() -> str 20 characters YYYYMMDDHHMMSSFFFFFF
# 0.23 = 2020-11-14 fixed n0list.compare()'s issue if more that one the same records are in the list
#                   *.compare() return ["othernotfound"] -> "self_unique"
#                   *.compare() return ["selfnotfound"]  -> "other_unique"
#                   *.compare() return ["notequal"]      -> "not_equal"
#                   get_composite_keys(..) -> generate_composite_keys(..)
# 0.24 = 2020-11-20 added is_exist(..), rewritten to_json(..), AddElem(..)'s optimization started
# 0.25 = 2020-11-20 removed:
#                       n0dict. __FindElem(..)
#                       n0dict. __AddElem(..)
#                       n0dict. AddElem(..)
#                   rewritten:
#                       n0dict. __getitem__(..)
#                       n0dict. __setitem__(..)
#                   added:
#                       n0dict. _find(..)
#                       n0dict. _add(..)
# 0.26 = 2020-12-15 rewritten n0dict. nvl()
# 0.27 = 2020-12-16
#                   added:
#                       n0dict. _get(..)
#                   rewritten:
#                       n0dict. __getitem__(..)
#                       n0dict. nvl()
#                       n0dict. _add()
#                   fixed:
#                       n0pretty()
# 0.28 = 2020-12-17
#                   renamed:
#                       n0dict. nvl() -> get(..)
#                   rewritten:
#                       n0dict. __setitem__(..)
#                       n0dict. _add()
# 0.29 = 2020-12-18
#                   added:
#                       n0list. any_*()
# 0.30 = 2020-12-20
#                   added:
#                       n0list. __contains__()  # something in n0list()
#                       n0dict. valid()
# 0.31 = 2020-12-20
#                   rewritten:
#                       n0dict. _find(..)
#                       split_name_index(..)
# 0.32 = 2020-12-21
#                   fixed:
#                       n0dict. _find(..): if not parent_node: => if parent_node is None:
#                                          because of parent_node could be '' as allowed value
# 0.33 = 2020-12-22
#                   enhanced:
#                       n0dict. _find(..): parent_node become mandatory argument
#                       n0dict. _get(..): support leading '?' in xpath
# 0.34 = 2020-12-24
#                   enhanced:
#                       n0list. __init__(..): option recursively:bool = True was added
#                       n0dict. __init__(..): option recursively:bool = True was added
# 0.35 = 2020-12-26
#                   enhanced:
#                       n0dict. __xml(..): nodes int, float support added
#                       n0dict. to_xml(..): multi-root support added
#                   enhanced version of xmltodict0121 was incapsulated till changes will be merged with main branch of xmltodic
#                   xmltodict0121 enhancement: automaticaly creation n0dict/n0list structure during XML import
# 0.36 = 2020-12-27
#                   xmltodict0121 was removed -- using strandard fuctionality of json and xmltodict: just only dict to n0dict will be automatic converted
#                   during loading xml/json and automatic conversion list into n0list use named parameter recursively=True in the costructor:
#                   my_n0dict = n0dict(json_txt, recursively=True)
#                   enhanced:
#                       n0dict. __init__(..)
#                       n0list. __init__(..)
#                       n0dict. __path(..)
#                       test_n0struct.py
# 0.37 = 2020-12-28
#                   fixed:
#                       n0dict. __xml(..)
# 0.38 = 2020-12-29
#                   enhanced:
#                       n0dict. __init__(..): option force_n0dict == None|!None was added, used ONLY for JSON text convertion
#                           force_n0dict == False/0/None => create [] (ordinary dict) nodes during JSON text convertion (json.loads)
#                           force_n0dict == True/1/any => create n0dict() nodes during JSON text convertion (json.loads)
#                           recursively == True => convert all list/dict nodes created during JSON text convertion (json.loads) into n0list/n0dict
#
#                       Performance results of some real code with JSON convertion:
#                           JSON_struct = n0dict(JSON_txt, recursively=True) => n0dict/n0list:
#                               36.889860 MB memory is used, 2.566520 seconds are taken for execution
#                           JSON_struct = n0dict(JSON_txt, recursively=False) => dict/list => JSON_subnode = n0dict(JSON_struct["node/subnode"]):
#                               36.945560 MB memory is used, 2.546720 seconds are taken for execution
#                           JSON_struct = n0dict(JSON_txt, force_n0dict=True) => n0dict/list:
#                               36.995180 MB memory is used, 2.576020 seconds are taken for execution
#
#                       Results are VERY strange, but they are true:
#                           Minimum memory usage: load as list/dict, and after convert all of them into n0list/n0dict inside constructor
#                           Maximum speed: load as list/dict, and after convert just requiered nodes into n0dict
#                       *** BEFORE MAKING DECISION MAKE YOUR OWN PERFORMACE TESTING ***
#
#                       Same for XML:
#                           XML_struct = n0dict(XML_txt, recursively=True) => n0dict/n0list:
#                           XML_struct = n0dict(XML_txt, recursively=False) => n0dict/list
#                           XML_struct = n0dict(XML_txt, force_n0dict=True) =>
#                               force_n0dict is ignored -- the same like n0dict(XML_txt, recursively=False) => n0dict/list
# 0.39 = 2021-01-04
#                   fixed:
#                       n0dict. __init__(..): xmltodict.parse(args[0], dict_constuctor = n0dict),
# 0.40 = 2021-01-08
#                   enhanced:
#                       n0pretty(item, indent_: int = 0, show_type:bool = True):
#                   added:
#                       n0dict. first(..)
#                       n0list. __getitem__(..)
#                       n0list. _get(..)
#                       n0list. get(..)
#                       n0list. first(..)
# 0.41 = 2021-01-09
#                   enhanced:
#                       Added predicate attrib[contains(text(),"TEXT")] or attrib[text()~~"TEXT"] or [attrib~~"TEXT"]
#                           xml.first('/repository/instanceInfo/instanceInfoProperty/@value[contains(text(),"PRE")]/../@value'))
#                           xml.first('/repository/instanceInfo/instanceInfoProperty/@value[text()~~"PRE"]/../@value'))
#                           xml.first('/repository/instanceInfo/instanceInfoProperty/[@value~~"PRE"]/@value'))
# 0.42 = 2021-01-22
#                   enhanced:
#                       n0debug(..)
#                       mypy optimization
# 0.43 = 2021-01-28
#                   enhanced:
#                       mypy optimization
#                       compare()["messages"] -> compare()["differences"]
#                       compare(exclude=) -> compare(exclude_xpaths=)
#                   added:
#                       def load_file(file_name: str) -> list:
#                       def save_file(file_name: str, lines: typing.Any):
#                       def load_serialized(...):
# 0.44 = 2021-02-08
#                   fixed:
#                       split_name_index(..)
# 0.45 = 2021-02-27
#                   enhanced:
#                       mypy optimization
#                       def init_logger(..)
#                   added:
#                       def timestamp() -> str:
#                       def date_yymmdd(now: typing.Union[datetime, None] = None, day_delta: int = 0, month_delta: int = 0) -> str:
#                       class OrderedSet(MutableSet):
#                       def unpack_references(initial_dict: dict, initial_key: str, recursive: bool = True) -> OrderedSet:
#                       class Git():
#                   fixed:
#                       n0list. _get(..)
# 0.45 = 2021-03-05
#                   added:
#                       n0dict. def update(self, xpath: typing.Union[dict, str], new_value: str = None) -> n0dict:
#                       n0dict. def delete(self, xpath: str, recursively: bool = False) -> n0dict:
#                       n0dict. def pop(self, xpath: str, recursively: bool = False) -> typing.Any:
# 0.46 = 2021-07-21 optimized for debugging
# 0.47 = 2021-07-21 fix for n0pretty()
# 0.48 = 2021-07-22
#                   added:
#                       def mask_number(not_masked_number: str):
#                       def unmask_number(masked_number: str):
#                       def mask_pan(buffer: str):
# 0.49 = 2021-07-23 fix for n0pretty()
# 0.50 = 2021-08-04 Impossible easely adopt for 3.6, only for 3.7, because of some modules (for example: immutables)
#                   are not precompiled for 3.6 at pypi.org. So installing of Visual C/C++ (or MinGW) is required.
# 0.51 = 2021-09-01
#                   fixed:
#                       n0dict. to_xpath(): prevent xpath generating for empty structures
#                       n0dict. _find(): changing incoming var name found_xpath_str -> xpath_found_str
# 0.52 = 2021-09-02
#                   added/enhanced:
#                       n0dict. __init__(file="{file_path}")
#                       n0dict. to_json/n0pretty(pairs_in_one_line=True)
# 0.53 = 2022-01-20 enhanced: n0dict. to_json/n0pretty(skip_none = False, skip_empty_arrays = False, skip_simple_types = True, auto_quotes = True)
# 0.54 = 2022-01-21 fixed defect got in 0.2.51, some enchancements in n0pretty
# 0.55 = 2022-01-21 try #2 to split single __init__.py into separate files
# 0.56 = 2022-01-23 fixed defect with using child class in parent
# 0.57 = 2022-01-24 added: 
#                       def is_date_format(date: datetime.date, format: str) -> bool:
#                       def load_ini(file_name: str, default_value = None, equal_sign = '=') -> dict:
#                       def generate_csv(root_node:n0dict, list_xpath:str, mapping_dict:dict, file_path:str = None, separator:str = '|') -> list:
# ******************************************************************************
# ******************************************************************************
from .n0struct_date import *
from .n0struct_random import *
from .n0struct_references import *
from .n0struct_git import *
from .n0struct_arrays import *
from .n0struct_mask import *
from .n0struct_utils import *
from .n0struct_transform_structure import *
from .n0struct_logging import *
from .n0struct_files import *
from .n0struct_files_csv import *
from .n0struct_files_fwf import *
from .n0struct_utils_find import *
from .n0struct_utils_compare import *
from .n0struct_findall import *
from .n0struct_n0list_n0dict import *
# ******************************************************************************
# ******************************************************************************



import typing
from pathlib import Path
import n0struct
# from .n0struct_logging import *
from .n0struct_n0list_n0dict import n0dict
# ******************************************************************************
# ******************************************************************************
def load_fwf_format(file_path: str):
    # fwf_format file is csv file, contains columns:
    #   #,name,offset,size,till
    loaded_format = load_csv(file_path)
    for i,column in enumerate(loaded_format):
        if not (column_offset := column['offset']) is None and column_offset != "":
            loaded_format[i]['offset'] = int(column_offset)
        else:
            loaded_format[i]['offset'] = None
        if not (column_till := column['till']) is None and column_till != "":
            loaded_format[i]['till']   = int(column_till)
        else:
            loaded_format[i]['till']   = None
        if not (column_size := column['size']) is None and column_size != "":
            loaded_format[i]['size']   = int(column_size)
        else:
            loaded_format[i]['size']   = None
    return loaded_format
# ******************************************************************************
def parse_fwf_line(incoming_row: str, fwf_format: dict, validate: bool = True, validate_empty: bool = True):
    if not fwf_format:
        raise Exception("fwf_format is mandatory parameter")
    parsed_line = {}
    for format_column in fwf_format:
        if format_column['name']:
            column_value = None
            if not (format_column_offset := format_column['offset']) is None and \
               not (format_column_till := format_column['till']) is None:
                column_value = incoming_row[format_column_offset:format_column_till].strip()
            parsed_line.update({format_column['name']: column_value})
    if validate:
        error_messages =  []
        for format_column in fwf_format:
            if (format_column_validation := format_column['validation']):
                lambda_function_for_validation = eval("lambda column_value, row: " + format_column_validation)
                if not lambda_function_for_validation(column_value, parsed_line):
                    error_messages.append(format_column['error_message'])
        if error_messages:
            return incoming_row, ";".join(error_messages)
            
    return parsed_line
# ******************************************************************************
def load_fwf(file_path:str, header_format: dict, body_format: dict = None, tail_format: dict = None, validate: bool = True, validate_empty: bool = True):
    success_parsed_lines = []
    failed_lines = []
    if not header_format:
        raise Exception("header_format is mandatory parameter")
    if not body_format:
        body_format = header_format
    if not tail_format:
        tail_format = body_format

    with open(file_path, 'rt') as filehandler:
        previous_line = None
        for i, line in enumerate(filehandler.read().split('\n')):
            if previous_line:
                parsed_line = parse_fwf_line(previous_line, header_format if i == 1 else body_format, validate, validate_empty)
                if isinstance(parsed_line, dict):
                    success_parsed_lines.append(parsed_line)
                else:
                    failed_lines.append(parsed_line)
            previous_line = line
        if previous_line:
            parsed_line = parse_fwf_line(previous_line, tail_format, validate, validate_empty)
            if isinstance(parsed_line, dict):
                success_parsed_lines.append(parsed_line)
            else:
                failed_lines.append(parsed_line)
    return success_parsed_lines, failed_lines
# ******************************************************************************
def generate_fwf_row(struct_to_save: dict, fwf_format: dict, filler:str = ' '):
    if not fwf_format:
        raise Exception("fwf_format is mandatory parameter")

    line_len = max([column['till'] for column in fwf_format])
    rendered_line = filler*line_len

    for format_column in fwf_format:
        format_column_mapping = None
        if format_column['name'] in struct_to_save:
            column_value = struct_to_save[format_column['name']]
        elif (format_column_mapping := format_column['mapping']):
            # n0struct.n0debug("struct_to_save")
            lambda_function_for_mapping = eval("lambda incoming_row: " + format_column_mapping)
            # n0struct.n0debug("format_column_mapping")
            column_value = lambda_function_for_mapping(struct_to_save)
        else:
            continue
        format_column_size = format_column['size']
        if format_column.get('type') == 'int':
            column_value = str(column_value).zfill(format_column_size)[:format_column_size]
        else:
            column_value = str(column_value).ljust(format_column_size)[:format_column_size]
        rendered_line = rendered_line[:format_column['offset']] +  column_value + rendered_line[format_column['till']:]

    return rendered_line
# ******************************************************************************
def generate_fwf(root_node:dict, list_xpath:str, mapping_dict:dict, fwf_format: dict, save_to:str = None, filler:str = ' ') -> list:
    if isinstance(root_node, (list, tuple)):
        list_of_items = root_node
    else:
        list_of_items = root_node.get(list_xpath, tuple())
        
    if save_to:
        out_filehandler = open(save_to, 'wt')

    fwf_table = []
    if list_of_items:
        if mapping_dict is None:
            if isinstance(list_of_items[0], list):
                header = [str(i) for i in range(0,len(list_of_items[0]))]
            elif isinstance(list_of_items[0], dict):
                header = list(list_of_items[0].keys())
            mapping_dict = {column_name: column_name for column_name in header}
        else:
            header = list(mapping_dict.keys())

        for found_item in list_of_items:  # found_item == row in case of CSV list or item node in case of XML structure
            if isinstance(found_item, dict):
                found_item = n0struct.n0dict(found_item) # to have ability to use .first(xpath)
            elif isinstance(found_item, list):
                found_item = n0struct.n0list(found_item) # to have ability to use .first(xpath)

            fwf_row = {}
            for column_name in mapping_dict:
                xpaths = mapping_dict[column_name]
                if not isinstance(xpaths, (list, tuple)):
                    xpaths = [xpaths]
                for xpath in xpaths:
                    found_value = found_item.first(xpath, "")
                    if found_value:
                        break
                fwf_row.update({column_name: found_value})
            fwf_table.append(fwf_row)

            if save_to:
                out_filehandler.write(generate_fwf_row(fwf_row, fwf_format, filler))

    if save_to:
        out_filehandler.close

    return fwf_table
# ******************************************************************************
# ******************************************************************************



import os
import sys
mydir = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, mydir)
sys.path.insert(0, mydir+"/../")
sys.path.insert(0, mydir+"/../../")
# import n0struct
# n0struct._DIFFTYPES = True  # For updating global variables using prefix is mandatory
# n0struct._DIFFVALUES = True
from n0struct import * # For using n0dict(), n0print(), n0debug without prefixes
set__flag_compare_check_different_types(True)
set__flag_compare_return_difference_of_values(True)
init_logger(debug_timeformat = None, debug_show_object_id = False, debug_logtofile = False)

# ******************************************************************************
# Etalon list in dictionary
# ******************************************************************************
dict1 = n0dict({
    "C": [
        {"a": 1, "b": 2, "c": 3, "value1": 1, "value2": 4},
        {"a": 4, "b": 5, "c": 6, "value1": 2, "value2": 5},
        {"a": 7, "b": 8, "c": 9, "value1": 3, "value2": 6},
    ],
}, recursively=True)
# ******************************************************************************
# Sorted list in dictionary
# ******************************************************************************
dict2 = n0dict({
    "C": [
        {"a": 1, "b": 2, "c": 3, "value1": 1, "value2": 4},
        {"a": 4, "b": 5, "c": 6, "value1": 2, "value2": 99},    # Single difference in value2
        {"a": 7, "b": 8, "c": 9, "value1": 3, "value2": 6},
    ],
}, recursively=True)
# ******************************************************************************
# Unsorted list in dictionary
# ******************************************************************************
dict3 = n0dict({
    "C": [
        {"a": 7, "b": 8, "c": 9, "value1": 3, "value2": 6},     # Changed order [2] -> [0]
        {"a": 4, "b": 5, "c": 6, "value1": 2, "value2": 99},    # Single difference in value2
        {"a": 1, "b": 2, "c": 3, "value1": 1, "value2": 4},     # Changed order [0] -> [2]
        {"a": 1, "b": 2, "c": 3, "value1": 1, "value2": 4},     # Duplicated with [2]
    ],
}, recursively=True)
# ******************************************************************************
def test_SortedLists():
    n0print("*"*80 + " 1 = Sorted list in dictionary = direct_compare")
    differences1_direct_compare = dict1.direct_compare(dict2, "dict1", "dict2")
    for key in differences1_direct_compare:
        n0print("*"*3 + " " + key)
        for itm in differences1_direct_compare[key]:
            n0print(itm)

    n0print("*"*80 + " 2 = Sorted list in dictionary = wise_compare")
    differences2_wise_compare = dict1.compare(dict2, "dict1", "dict2",
        composite_key = ("a", "b", "c"),
        compare_only = ("value1", "value2")
    )
    for key in differences2_wise_compare:
        n0print("*"*3 + " " + key)
        for itm in differences2_wise_compare[key]:
            n0print(itm)

    n0print("="*80)
    n0debug("differences1_direct_compare")
    n0debug_calc(notemptyitems(differences1_direct_compare["differences"]), 'notemptyitems(differences1_direct_compare["differences"])')
    assert notemptyitems(differences1_direct_compare["differences"]) == 1
    n0debug_calc(notemptyitems(differences1_direct_compare["not_equal"]), 'notemptyitems(differences1_direct_compare["not_equal"])')
    assert notemptyitems(differences1_direct_compare["not_equal"]) == 4
    assert notemptyitems(differences1_direct_compare["difftypes"]) == 0
    assert notemptyitems(differences1_direct_compare["other_unique"]) == 0
    assert notemptyitems(differences1_direct_compare["self_unique"]) == 0

    n0print("="*80)
    n0debug("differences2_wise_compare")
    assert differences1_direct_compare["differences"]     == differences2_wise_compare["differences"]
    assert differences1_direct_compare["not_equal"]     == differences2_wise_compare["not_equal"]
    assert differences1_direct_compare["difftypes"]    == differences2_wise_compare["difftypes"]
    assert differences1_direct_compare["other_unique"] == differences2_wise_compare["other_unique"]
    assert differences1_direct_compare["self_unique"]== differences2_wise_compare["self_unique"]
# ******************************************************************************
def test_UnsortedLists():
    n0print("*"*80 + " 3 = Unsorted list in dictionary = direct_compare")
    differences1_direct_compare = dict1.direct_compare(dict3, "dict1", "dict3")
    for key in differences1_direct_compare:
        n0print("*"*3 + " " + key)
        for itm in differences1_direct_compare[key]:
            n0print(itm)

    n0print("*"*80 + " 4 = Unsorted list in dictionary = compare")
    differences2_wise_compare = dict1.compare(dict3, "dict1", "dict3",
        composite_key = ("a", "b", "c"),
        compare_only = ("value1", "value2")
    )
    for key in differences2_wise_compare:
        n0print("*"*3 + " " + key)
        for itm in differences2_wise_compare[key]:
            n0print(itm)

    n0print("="*80)
    n0debug("differences1_direct_compare")
    n0debug_calc(notemptyitems(differences1_direct_compare["differences"]), 'notemptyitems(differences1_direct_compare["differences"]')
    assert notemptyitems(differences1_direct_compare["differences"]) == 12
    n0debug_calc(notemptyitems(differences1_direct_compare["not_equal"]), 'differences1_direct_compare["not_equal"]')
    assert notemptyitems(differences1_direct_compare["not_equal"]) == 44
    assert notemptyitems(differences1_direct_compare["difftypes"]) == 0
    n0debug_calc(notemptyitems(differences1_direct_compare["other_unique"]), 'notemptyitems(differences1_direct_compare["other_unique"])')
    assert notemptyitems(differences1_direct_compare["other_unique"]) == 6
    n0debug_calc(notemptyitems(differences1_direct_compare["self_unique"]), 'notemptyitems(differences1_direct_compare["self_unique"])')
    assert notemptyitems(differences1_direct_compare["self_unique"]) == 0

    n0print("="*80)
    n0debug("differences2_wise_compare")
    n0debug_calc(notemptyitems(differences2_wise_compare["differences"]), 'notemptyitems(differences2_wise_compare["differences"]')
    assert notemptyitems(differences2_wise_compare["differences"]) == 2
    n0debug_calc(notemptyitems(differences2_wise_compare["not_equal"]), 'differences2_wise_compare["not_equal"]')
    assert notemptyitems(differences2_wise_compare["not_equal"]) == 4
    assert notemptyitems(differences2_wise_compare["difftypes"]) == 0
    n0debug_calc(notemptyitems(differences2_wise_compare["other_unique"]), 'notemptyitems(differences2_wise_compare["other_unique"])')
    assert notemptyitems(differences2_wise_compare["other_unique"]) == 6
    n0debug_calc(notemptyitems(differences2_wise_compare["self_unique"]), 'notemptyitems(differences2_wise_compare["self_unique"])')
    assert notemptyitems(differences2_wise_compare["self_unique"]) == 0
# ******************************************************************************
def main():
    n0print("="*80)
    n0debug_calc(dict1,"dict1")
    n0print("="*80)
    n0debug_calc(dict2,"dict2")
    n0print("="*80)
    n0debug_calc(dict2,"dict3")

    test_SortedLists()
    test_UnsortedLists()

    n0debug_calc(dict1,'dict1')
    dict1["moreone/node[new()]/value"] = n0dict()
    n0debug_calc(dict1,'dict1')
    dict1["moreone/node[new()]/value"] = None
    n0debug_calc(dict1,'dict1')  # Defect: new() has not converted single n0dict node into node[]

    dict1["moreone/node[new()]/value"] = ""
    dict1["moreone/node[new()]/value"] = 1
    dict1["moreone/node[last()]/value"] = 2
    dict1["moreone/node[last()]/code"] = "two"
    dict1["moreone/node[new()]/value"] = 3
    dict1["moreone/node[-1]/value"] = 4
    dict1["moreone/node[-1]/code"] = "four"

    assert isinstance(dict1["moreone/node[0]/value"],n0dict) == True
    assert (dict1["moreone/node[1]/value"] is None) == True
    assert isinstance(dict1["moreone/node[2]/value"],str) == True
    assert (not dict1["moreone/node[2]/value"]) == True
    assert dict1["moreone/node[3]/value"] == 2
    assert dict1["moreone/node[4]/value"] == 4

    n0print(dict1.to_json())
    n0print(dict1.to_xml())

    n0print(dict1["moreone/node"])

    assert len(dict1["moreone/node"]) == 5
    assert len(dict1["moreone/node[*]"]) == 5

    n0print(dict1["moreone"])
    assert len(dict1["moreone"]) == 1

    n0print(dict1["moreone/node/code[text()='four']/../value"])
    assert dict1["moreone/node/code[text()='four']/../value"][0] == 4
    assert dict1["moreone/node/code[text()='four']/../value"] == [4]


    n0print(dict1["moreone/node/code[text()='two']/../value"])
    assert dict1["moreone/node/code[text()='two']/../value"] == [2]

    dict1["moreone/node/code[text()='two']/../value"] = 6
    n0print(dict1["moreone/node/code[text()='two']/../value"])
    assert dict1["moreone/node/code[text()='two']/../value"] == [6]
    assert dict1["moreone/node/code[text()=two]/../value"] == [6]
    assert dict1['moreone/node/code[text()="two"]/../value'] == [6]

    n0print(dict1["moreone/node[code='two']/value"])
    assert dict1["moreone/node[code='two']/value"] == [6]

    dict1["moretwo/node/code"] = "seven"
    dict1["moretwo/node/value"] = 7
    n0print(dict1["moretwo/node/code[text()='seven']/../value"])
    assert dict1["moretwo/node/code[text()='seven']/../value"] == 7

    n0print(dict1["moretwo/node[code='seven']/value"])
    assert dict1["moretwo/node[code='seven']/value"] == 7

    n0print(dict1.get("moretwo/node/code[text()='eight']/../value"))
    assert dict1.get("moretwo/node/code[text()='eight']/../value") is None

    n0print(dict1["?moretwo/node/code[text()='eight']/../value"] or None)
    assert (dict1["?moretwo/node/code[text()='eight']/../value"] or None) is None

    # set__debug_output(sys.stdout.write)

    str_json = """
    {
        "root": {
            "a": 1.0,
            "b": {
                "c": 2,
                "d": {
                    "e": {
                        "f": 3
                    },
                    "g": [
                        "4",
                        {"h": 5}
                    ],
                    "i": [
                        {"j": 6}
                    ]
                }
            },
            "k": [
                7,
                8
            ]
        }
    }
    """
    dict_from_json_default = n0dict(str_json)
    n0debug("dict_from_json_default")
    assert isinstance(dict_from_json_default, n0dict) == True
    assert isinstance(dict_from_json_default["root"], dict) == True
    assert isinstance(dict_from_json_default["root/a"], float) == True # XML supports only str, for JSON float
    assert isinstance(dict_from_json_default["root/b"], dict) == True
    assert isinstance(dict_from_json_default["root/b/d"], dict) == True
    assert isinstance(dict_from_json_default["root/b/d/e"], dict) == True
    assert isinstance(dict_from_json_default["root/b/d/g"], list) == True # For recursively=False
    assert isinstance(dict_from_json_default["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_json_default["root/b/d/g[1]"], dict) == True
    assert isinstance(dict_from_json_default["root/b/d/g[1]/h"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_default["root/b/d/i"], list) == True # For JSON list
    assert isinstance(dict_from_json_default["root/b/d/i[0]"], dict) == True
    assert isinstance(dict_from_json_default["root/b/d/i[0]/j"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_default["root/k"], list) == True # For recursively=False
    assert isinstance(dict_from_json_default["root/k[0]"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_default["root/k[1]"], int) == True # XML supports only str, for JSON int
    dict_from_json_default.to_json(); dict_from_json_default.to_xml(); dict_from_json_default.to_xpath()

    dict_from_json_force_n0dict = n0dict(str_json, force_n0dict = True)
    n0debug("dict_from_json_force_n0dict")
    assert isinstance(dict_from_json_force_n0dict, n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root"], n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root/a"], float) == True # XML supports only str, for JSON float
    assert isinstance(dict_from_json_force_n0dict["root/b"], n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root/b/d"], n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root/b/d/g"], list) == True # For recursively n0list
    assert isinstance(dict_from_json_force_n0dict["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_json_force_n0dict["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root/b/d/g[1]/h"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_force_n0dict["root/b/d/i"], list) == True # For JSON list, for JSON recursively n0list
    assert isinstance(dict_from_json_force_n0dict["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_json_force_n0dict["root/b/d/i[0]/j"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_force_n0dict["root/k"], list) == True # For recursively n0list
    assert isinstance(dict_from_json_force_n0dict["root/k[0]"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_force_n0dict["root/k[1]"], int) == True # XML supports only str, for JSON int
    dict_from_json_force_n0dict.to_json(); dict_from_json_force_n0dict.to_xml(); dict_from_json_force_n0dict.to_xpath()

    dict_from_json_recursevely = n0dict(str_json, recursively=True)
    n0debug("dict_from_json_recursevely")
    assert isinstance(dict_from_json_recursevely, n0dict) == True
    assert isinstance(dict_from_json_recursevely["root"], n0dict) == True
    assert isinstance(dict_from_json_recursevely["root/a"], float) == True # XML supports only str, for JSON float
    assert isinstance(dict_from_json_recursevely["root/b"], n0dict) == True
    assert isinstance(dict_from_json_recursevely["root/b/d"], n0dict) == True
    assert isinstance(dict_from_json_recursevely["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_json_recursevely["root/b/d/g"], n0list) == True # For recursively=True
    assert isinstance(dict_from_json_recursevely["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_json_recursevely["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_json_recursevely["root/b/d/g[1]/h"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_recursevely["root/b/d/i"], n0list) == True # For JSON list, for JSON recursively n0list
    assert isinstance(dict_from_json_recursevely["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_json_recursevely["root/b/d/i[0]/j"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_recursevely["root/k"], n0list) == True # For recursively=True
    assert isinstance(dict_from_json_recursevely["root/k[0]"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_recursevely["root/k[1]"], int) == True # XML supports only str, for JSON int
    dict_from_json_recursevely.to_json(); dict_from_json_recursevely.to_xml(); dict_from_json_recursevely.to_xpath()

    dict_from_json_recursevely_force_n0dict = n0dict(str_json, recursively=True, force_n0dict = True)
    n0debug("dict_from_json_recursevely_force_n0dict")
    assert isinstance(dict_from_json_recursevely_force_n0dict, n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root"], n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/a"], float) == True # XML supports only str, for JSON float
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b"], n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d"], n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/g"], n0list) == True # For recursively=True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/g[1]/h"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/i"], n0list) == True # For JSON list, for JSON recursively n0list
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/b/d/i[0]/j"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/k"], n0list) == True # For recursively=True
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/k[0]"], int) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_json_recursevely_force_n0dict["root/k[1]"], int) == True # XML supports only str, for JSON int
    dict_from_json_recursevely_force_n0dict.to_json(); dict_from_json_recursevely_force_n0dict.to_xml(); dict_from_json_recursevely_force_n0dict.to_xpath()

    str_xml = """
    <root>
        <a>1</a>
        <b>
            <c>2</c>
            <d>
                <e>
                    <f>3</f>
                </e>
                <g>
                    4
                </g>
                <g>
                    <h>5</h>
                </g>
                <i>
                    <j>6</j>
                </i>
            </d>
        </b>
        <k>7</k>
        <k>8</k>
    </root>
    """
    dict_from_xml_default = n0dict(str_xml)
    n0debug("dict_from_xml_default")
    assert isinstance(dict_from_xml_default, n0dict) == True
    assert isinstance(dict_from_xml_default["root"], n0dict) == True
    assert isinstance(dict_from_xml_default["root/a"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_default["root/b"], n0dict) == True
    assert isinstance(dict_from_xml_default["root/b/d"], n0dict) == True
    assert isinstance(dict_from_xml_default["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_xml_default["root/b/d/g"], list) == True # For recursively=False
    assert isinstance(dict_from_xml_default["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_xml_default["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_xml_default["root/b/d/g[1]/h"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_default["root/b/d/i"], n0dict) == True # For JSON list, for JSON recursively n0list
    assert isinstance(dict_from_xml_default["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_xml_default["root/b/d/i[0]/j"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_default["root/k"], list) == True # For recursively=False
    assert isinstance(dict_from_xml_default["root/k[0]"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_default["root/k[1]"], str) == True # XML supports only str, for JSON int
    dict_from_xml_default.to_json(); dict_from_xml_default.to_xml(); dict_from_xml_default.to_xpath()

    # force_n0dict = True for XML is useless, in any cases dictionaries will be n0dict
    dict_from_xml_force_n0dict = n0dict(str_xml, force_n0dict = True)
    n0debug("dict_from_xml_force_n0dict")
    assert isinstance(dict_from_xml_force_n0dict, n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root"], n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root/a"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_force_n0dict["root/b"], n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root/b/d"], n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/g"], list) == True # For recursively=False
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/g[1]/h"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/i"], n0dict) == True # For JSON list
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_xml_force_n0dict["root/b/d/i[0]/j"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_force_n0dict["root/k"], list) == True # For recursively=False
    assert isinstance(dict_from_xml_force_n0dict["root/k[0]"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_force_n0dict["root/k[1]"], str) == True # XML supports only str, for JSON int
    dict_from_xml_force_n0dict.to_json(); dict_from_xml_force_n0dict.to_xml(); dict_from_xml_force_n0dict.to_xpath()

    dict_from_xml_recursevely = n0dict(str_xml, recursively=True)
    n0debug("dict_from_xml_recursevely")
    assert isinstance(dict_from_xml_recursevely, n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root/a"], str) == True # XML supports only str
    assert isinstance(dict_from_xml_recursevely["root/b"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d/g"], n0list) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d/g[1]/h"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_recursevely["root/b/d/i"], n0dict) == True # For JSON list
    assert isinstance(dict_from_xml_recursevely["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely["root/b/d/i[0]/j"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_recursevely["root/k"], n0list) == True
    assert isinstance(dict_from_xml_recursevely["root/k[0]"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_recursevely["root/k[1]"], str) == True # XML supports only str, for JSON int
    dict_from_xml_recursevely.to_json(); dict_from_xml_recursevely.to_xml(); dict_from_xml_recursevely.to_xpath()

    dict_from_xml_recursevely_force_n0dict = n0dict(str_xml, recursively=True, force_n0dict = True)
    n0debug("dict_from_xml_recursevely_force_n0dict")
    assert isinstance(dict_from_xml_recursevely_force_n0dict, n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/a"], str) == True # XML supports only str
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/e"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/g"], n0list) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/g[0]"], str) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/g[1]"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/g[1]/h"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/i"], n0dict) == True # For JSON list
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/i[0]"], n0dict) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/b/d/i[0]/j"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/k"], n0list) == True
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/k[0]"], str) == True # XML supports only str, for JSON int
    assert isinstance(dict_from_xml_recursevely_force_n0dict["root/k[1]"], str) == True # XML supports only str, for JSON int
    dict_from_xml_recursevely_force_n0dict.to_json(); dict_from_xml_recursevely_force_n0dict.to_xml(); dict_from_xml_recursevely_force_n0dict.to_xpath()

    str_xml_empty = "<root/>"
    dict_from_xml_empty = n0dict(str_xml_empty)
    n0debug("dict_from_xml_empty")
    assert isinstance(dict_from_xml_empty, n0dict) == True
    assert (dict_from_xml_empty["root"] is None) == True
    dict_from_xml_empty.to_json(); dict_from_xml_empty.to_xml(); dict_from_xml_empty.to_xpath()

    dict4 = n0dict().update(
        {
            "a": 1,
            "b": {
                "c": 2,
                "d": {
                    "e": 3
                }
            }
        }
    )
    n0debug("dict4")
    assert isinstance(dict4, n0dict) == True
    assert dict4["a"] == 1
    assert dict4["b/c"] == 2
    assert dict4["b/d/e"] == 3
    assert dict4.get("b/d/e", "ALREADY_DELETED") == 3
    assert dict4.pop("b/d/e") == 3
    assert dict4.get("b/d/e", "ALREADY_DELETED") == "ALREADY_DELETED"
    assert dict4.delete("b/c")["a"] == 1
    assert dict4.get("b/c", "ALREADY_DELETED") == "ALREADY_DELETED"
    assert dict4.delete("b/d", recursively=True)
    assert dict4.get("b", "ALREADY_DELETED") == "ALREADY_DELETED"
    n0debug("dict4")

    dict5 = n0dict().update(
        {
            "a": {
                "b": {
                    "c": {
                        "d": 4
                    }
                }
            }
        }
    )
    n0debug("dict5")
    dict5.delete("a/b/c/d", recursively=True)
    n0debug("dict5")

    # ##########################################################################
    # n0struct_findall.py
    # ##########################################################################
    dict6 = n0dict({
        "Root": {
            "Node1": {
                "Subnode1_1": [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode1_1"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode1_1"},
                ],
                "Subnode1_2": [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode1_2"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode1_2"},
                ],
            },
            "Node2": {
                "Subnode2_1": [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode2_1"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode2_1"},
                ],
                "Subnode2_2": [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode2_2"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode2_2"},
                ],
            },
        }
    })
    n0debug("dict6")
    n0debug_calc(dict6.findall("//Root/Node1"))
    n0debug_calc(dict6.findall("//Root/Node2/Subnode2_1/tag[text()==PARAM1]/../value"))
    assert list(dict6.findall("//Root/Node2/Subnode2_1/tag[text()==PARAM1]/../value").keys())[0] == "//Root/Node2/Subnode2_1[0]/value"
    assert list(dict6.findall("//Root/Node2/Subnode2_1/tag[text()==PARAM1]/../value").values())[0] == "VALUE1 from subnode2_1"

    dict7 = n0dict({
        "Root": {
            "Nodes": [
                [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode0_0"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode0_1"},
                ],
                [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode1_0"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode1_1"},
                ],
                [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode2_0"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode2_1"},
                ],
                [
                    {"tag": "PARAM1", "value": "VALUE1 from subnode3_0"},
                    {"tag": "PARAM2", "value": "VALUE2 from subnode3_1"},
                ],
            ],
        }
    })
    n0print("\n" + list(dict7.findall("//").values())[0].to_xpath())
    found = dict7.findall("//Root/Nodes/tag[text()==PARAM1]/../value")
    n0debug_calc(found, '"//Root/Nodes/tag[text()==PARAM1]/../value"')
    assert len(found) == 4
    assert list(found.values())[2] == "VALUE1 from subnode2_0"

if __name__ == '__main__':
    main()
    n0print("Mission acomplished")



import typing
# ******************************************************************************
# ******************************************************************************
def n0isnumeric(value: str) -> bool:
    # return value.translate(str.maketrans("+-.", "000")).isnumeric() # Py3 dirty fix
    if isinstance(value, (int, float)):
        return True
    if not isinstance(value, str):
        return False
    value = value.strip()
    if value.startswith('+') or value.startswith('-'):
        value = value[1:].strip()
    if len(value) > 10:
        return False
    if value.count('.') == 1:
        value = value.replace('.','0')
    return value.isnumeric()
# ******************************************************************************
def get_key_by_value(dict_: dict, value_: typing.Any):
    """
    :param dict_:
    :param value_:
    :return: last key which is associated with value_ in dict_
    """
    return {value: key for key, value in dict_.items()}[value_]
# ******************************************************************************
def n0eval(_str: str) -> typing.Union[int, float, typing.Any]:
    def my_split(_str: str, _separator: str) -> typing.List:
        return [
                (_separator if _separator != '+' and i else "") + itm.strip()
                for i, itm in enumerate(_str.split(_separator))
                if itm.strip()
        ]

    if not isinstance(_str, str):
        return _str

    _str = _str.replace(" ","").lower()
    if not _str:
        return _str
        # raise ValueError("Could not convert empty/null string into index")

    first_split = my_split(_str, '+')
    second_split = []
    for item in first_split:
        items = my_split(item, '-')
        second_split.extend(items)

    result = 0
    for item in second_split:
        if item == "new()":
            return _str
        if item == "last()":
            item = -1
        else:
            try:
                if '.' in item:
                    item = float(item)
                else:
                    item = int(item)
            except Exception:
                return _str
        result += item

    return result
# ******************************************************************************
def raise_in_lambda(ex): raise ex
# ******************************************************************************
def deserialize_list(
                        buffer: str,
                        separator_tag: str = ",",
                        process_item = (lambda item: item),
) -> list:
    if not buffer:
        return []
    return [
            process_item(stripped_item)
            for item in buffer.split(separator)
            if (stripped_item := item.strip())
    ]
# ******************************************************************************
def deserialize_list_of_lists(
                        buffer: str,
                        separator_tag: str = ",",
                        process_item = (lambda item: item),
) -> list:
    return deserialize_list(buffer, ';', lambda item: deserialize_list(item, ',', process_item))
# ******************************************************************************
def create_fixed_list(
                        buffer: str,
                        dst_list_len: int,
                        default_value: int = None,
                        separator_tag: str = ",",
                        process_item = (lambda item: item),
) -> list:
    '''
    generate list [value1, value2, ... valueN] with size of dst_list_len
    from deserialized buffer
    in case of value is not existed, then [value1, default_value, ... default_value]
    '''
    src_list = deserialize_list(buffer, separator_tag)
    src_list_len = len(src_list)
    return [
        (
            (
                item
                if (item:=src_list[i])
                else default_value
            )
            if i < src_list_len
            else default_value
        )
        for i in range(dst_list_len)
    ] 
# ******************************************************************************
def deserialize_dict(
                        buffer: str,
                        separator_tag: str = ";",
                        equal_tag: str = "=",
) -> dict:
    if not isinstance(buffer, str) or not buffer:
        return {}

    return {
        (tag_value:=pair.split(equal_tag, 1))[0]: (tag_value[1] if len(tag_value) > 1 else "")
        for pair in buffer.split(separator_tag)     # split pairs by ';' in 'tag1=value1;tag2=value2'
        if pair                                     # mitigate ';;;' in 'tag1=value1;;;tag2=value2'
    }
# ******************************************************************************
def get_value_by_tag(
                        tag_name: str,
                        buffer: str,
                        default_value: typing.Any = None,
                        separator_tag: str = ";",
                        equal_tag: str = "=",
) -> str:
    if not isinstance(tag_name, str) or not tag_name:
        return default_value

    return deserialize_dict(buffer, separator_tag, equal_tag).get(tag_name, default_value)
# ******************************************************************************
def validate_str(value, default_value: str = "") -> str:
    return f"{default_value if not value else value}"
# ******************************************************************************
def validate_path(value, default_value: str = "") -> str:
    if not value:
        return default_value
    if value == ".":
        return ""
    return value
# ******************************************************************************
def validate_and_map(value, variants: dict, default_value: str = ""):
    if not value:
        return default_value
    if (upper_value:=value.upper()) in variants:
        return variants[upper_value]
    return default_value
# ******************************************************************************
def validate_bool(value, default_value: False):
    return validate_and_map(
                            value,
                            {
                                "TRUE": True, 
                                "YES": True, 
                                "Y": True, 
                                "1": True, 
                            },
                            default_value
    )
# ******************************************************************************
def validate_values(value: str, possible_values_the_last_is_default: typing.Union[list, tuple], default_value: str = ""):
    if not isinstance(possible_values_the_last_is_default, (list,tuple)) and not len(possible_values_the_last_is_default):
        return default_value
    if (upper_value:=(value or "").upper()) in possible_values_the_last_is_default:
        return upper_value 
    return possible_values_the_last_is_default[-1]
# ******************************************************************************
# ******************************************************************************



# from typing import Any, Union, Dict, Tuple, List, Set, FrozenSet, NewType, Sequence
import typing
# from mypy_extensions import (Arg, DefaultArg, NamedArg, DefaultNamedArg, VarArg, KwArg)
# from mypy_extensions import Arg

# from datetime import datetime, timedelta, date
import datetime

from .n0struct_logging import *

# ******************************************************************************
# ******************************************************************************
def date_today() -> datetime.datetime:
    """
    :return: today -> date

    https://stackoverflow.com/questions/32517248/what-is-the-difference-between-python-functions-datetime-now-and-datetime-t
        datetime.datetime.now() takes tzinfo as keyword argument but datetime.today() does not take any keyword arguments.
        By default, now() executes with datetime.datetime.now(tz=None)
    """
    return datetime.datetime.today()
# ******************************************************************************
def date_only(input_date: datetime.datetime) -> datetime.date:
    return input_date.date()
# ******************************************************************************
def date_delta(input_date: typing.Union[datetime.datetime, datetime.date, None] = None, day_delta: int = 0, month_delta: int = 0) -> datetime.datetime:
    """
    :param day_delta:
    :param month_delta:
    :return: today + day_delta + month_delta -> date
    """
    if not isinstance(input_date, datetime.datetime):
        if isinstance(input_date, datetime.date):
            input_date = datetime.datetime.combine(input_date, datetime.datetime.min.time())
        else:
            return None

    # n0debug("input_date")
    date_delta_ = input_date + datetime.timedelta(days=day_delta)
    # n0debug("date_delta_")
    
    month_quotient, month_remainder = divmod(date_delta_.month + month_delta - 1, 12)
    # n0debug("month_quotient")
    # n0debug("month_remainder")
    
    result = datetime.datetime(
                            date_delta_.year + month_quotient, month_remainder + 1, date_delta_.day,
                            date_delta_.hour, date_delta_.minute,  date_delta_.second,  date_delta_.microsecond
    )
    # n0debug("result")
    return result

# ******************************************************************************
def date_to_format(input_date: typing.Union[datetime.datetime, None], date_format: str, day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param date_format:
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> date_format
    """
    # n0debug("input_date")
    # n0debug("day_delta")
    # n0debug("month_delta")
    result = date_delta(input_date, day_delta, month_delta)
    # n0debug("result")
    # n0debug("date_format")
    if result:
        result = result.strftime(date_format)
    # n0debug("result")
    return result
# ******************************************************************************
# LEGACY
# ******************************************************************************
def date_format(date_format: str, input_date: typing.Union[datetime.datetime, None] = None, day_delta: int = 0, month_delta: int = 0) -> str:
    """
    LEGACY
    """
    return date_to_format(input_date or date_today(), date_format, day_delta, month_delta)
# ******************************************************************************
def date_timestamp_full(input_date: typing.Union[datetime.datetime, None] = None, day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str 20 characters YYYYMMDDHHMMSSFFFFFF
    """
    if not input_date:
        input_date = date_today()
    return date_to_format(input_date, "%Y%m%d%H%M%S%f", day_delta, month_delta)
# ******************************************************************************
def date_timestamp(input_date: typing.Union[datetime.datetime, None] = None) -> str:
    """
    :return: input_date -> str 13 characters YYMMDD_HHMMSS
    """
    timestamp = date_timestamp_full(input_date)
    return timestamp[2:8] + "_" + timestamp[8:14]
# ******************************************************************************
def date_iso(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str ISO date format
    """
    return date_delta(input_date, day_delta, month_delta).isoformat(timespec='microseconds')
# ******************************************************************************
def date_yymmdd(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str YYMMDD
    """
    return date_to_format(input_date, "%y%m%d", day_delta, month_delta)
# ******************************************************************************
def date_dash_yyyymmdd(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str YYYY-MM-DD
    """
    return date_to_format(input_date, "%Y-%m-%d", day_delta, month_delta)
# ******************************************************************************
def date_slash_ddmmyyyy(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str DD/MM/YYYY
    """
    return date_to_format(input_date, "%d/%m/%Y", day_delta, month_delta)
# ******************************************************************************
def date_dash_ddmmyyyy(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str DD-MM-YYYY
    """
    return date_to_format(input_date, "%d-%m-%Y", day_delta, month_delta)
# ******************************************************************************
def date_yymm(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str YYMM
    """
    return date_to_format(input_date, "%y%m", day_delta, month_delta)
# ******************************************************************************
def date_mmyy(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str MMYY
    """
    return date_to_format(input_date, "%m%y", day_delta, month_delta)
# ******************************************************************************
def date_julian(input_date: typing.Union[datetime.datetime, None], day_delta: int = 0, month_delta: int = 0) -> str:
    """
    :param input_date:
    :param day_delta:
    :param month_delta:
    :return: (input_date or today) + day_delta + month_delta -> str JULIAN
    """
    return date_to_format(input_date, "%j", day_delta, month_delta).zfill(3)
# ******************************************************************************
def time_hhmmss(input_date: typing.Union[datetime.datetime, None]) -> str:
    """
    :param input_date:
    :return: (input_date or today) -> str HHMMSS
    """
    return date_to_format(input_date, "%H%M%S")
# ******************************************************************************
def time_colon_hhmmss(input_date: typing.Union[datetime.datetime, None]) -> str:
    """
    :param input_date:
    :return: (input_date or today) -> str HHMMSS
    """
    return date_to_format(input_date, "%H:%M:%S")
# ******************************************************************************
def to_date(input_date_str: str, date_format: str = None) -> typing.Union[datetime.datetime, str]:
    """
    :param input_date_str:
    :param date_format:
    :return: input_date_str converted into date_time as date_format
    :return: str -> date
    """
    if not input_date_str:
        return input_date_str
    elif not isinstance(input_date_str, str):
        raise TypeError("{input_date_str=} must be str")

    if date_format:
        try:
            '''
            # result = datetime.datetime.strptime(input_date_str, date_format).date()
            result = datetime.datetime.strptime(input_date_str, date_format)
            ##n0debug_calc(result, f"to_date({input_date_str=}, {date_format=})")
            return result
            '''
            return datetime.datetime.strptime(input_date_str, date_format)
        except (ValueError, TypeError):
            ##n0debug_calc(input_date_str, f"FAILED: to_date({input_date_str=}, {date_format=})")
            return input_date_str
    else:
        try:
            return datetime.datetime.strptime(input_date_str, "%Y-%m-%d").date()  # 2020-07-16
        except (ValueError, TypeError):
            try:
                return datetime.datetime.strptime(input_date_str, "%d-%b-%y").date()  # 16-JUL-20
            except (ValueError, TypeError):
                try:
                    return datetime.datetime.strptime(input_date_str, "%d-%m-%Y").date()  # 16-07-2020
                except (ValueError, TypeError):
                    try:
                        return datetime.datetime.strptime(input_date_str, "%d.%m.%Y").date()  # 16.07.2020
                    except (ValueError, TypeError):
                        try:
                            return datetime.datetime.strptime(input_date_str, "%m/%d/%Y").date()  # 07/16/2020
                        except (ValueError, TypeError):
                            return input_date_str
# ******************************************************************************
# LEGACY
# ******************************************************************************
def from_date(input_date_str: str, date_format: str) -> typing.Union[datetime.datetime, str, None]:
    """
    LEGACY
    """
    return to_date(input_date_str, date_format)
# ******************************************************************************
def from_ddmmmyy(input_date_str: str) -> typing.Union[datetime.datetime, str, None]:
    """
    :param input_date_str: DD-MMM-YY # 16-JUL-20
    :return: str -> date
    """
    return to_date(input_date_str, "%d-%b-%y")  # 16-JUL-20
# ******************************************************************************
def from_yyyymmdd(input_date_str: str) -> typing.Union[datetime.datetime, str, None]:
    """
    :param input_date_str: YYYY-MM-DD # 2020-07-16
    :return: str -> date
    """
    return to_date(input_date_str, "%Y-%m-%d")  # 2020-07-16
# ******************************************************************************
def from_ddmmyyyy(input_date_str: str) -> typing.Union[datetime.datetime, str, None]:
    """
    :param input_date_str: DD-MM-YYYY # 16-07-2020
    :return: str -> date
    """
    return to_date(input_date_str, "%d-%m-%Y")  # 16-07-2020
# ******************************************************************************
def is_date_format(input_date_str: str, date_format: str) -> typing.Union[datetime.datetime, bool]:
    result = to_date(input_date_str, date_format)
    if isinstance(result, datetime.datetime):
        return result
    else:
        return False
# ******************************************************************************
def is_date_yymm(input_date_str: str) -> typing.Union[datetime.datetime, bool]:
    return is_date_to_format(input_date_str, "%y%m")
# ******************************************************************************
def is_date_mmyy(input_date_str: str) -> typing.Union[datetime.datetime, bool]:
    return is_date_to_format(input_date_str, "%m%y")
# ******************************************************************************
def first_day_of_yymm(input_date_str: str) -> typing.Union[datetime.datetime, bool]:
    return is_date_yymm(input_date_str)
# ******************************************************************************
def last_day_of_yymm(input_date_str: str) -> typing.Union[datetime.datetime, bool]:
    if not (date_datetime := is_date_yymm(input_date_str)):
        return date_datetime
    return date_delta(date_datetime, month_delta=1) - datetime.timedelta(microseconds=1)
# ******************************************************************************
# ******************************************************************************



import typing
from pathlib import Path
import n0struct
# from .n0struct_logging import *
from .n0struct_n0list_n0dict import n0dict
# ******************************************************************************
# ******************************************************************************
def parse_complex_csv_line(
    line:str,
    separator:str = ',',
    strip_column:callable = lambda column_value: column_value,
) -> list:
    fields_in_the_row = []
    flag_delimiter_in_field = flag_expect_delimiter_or_quotes = False
    field_value = ""
    for i,ch in enumerate(line):
        if ch == separator and (not flag_delimiter_in_field or flag_expect_delimiter_or_quotes):
            fields_in_the_row.append(strip_column(field_value))
            flag_delimiter_in_field = flag_expect_delimiter_or_quotes = False
            field_value = ""
            continue  # skip separator
        elif ch == '"':
            if len(field_value) == 0:
                flag_delimiter_in_field = True
                continue  # skip " in the begining of the field
            elif flag_delimiter_in_field:
                if not flag_expect_delimiter_or_quotes:
                    flag_expect_delimiter_or_quotes = True
                    continue  # skip first " in the middle of line
                flag_expect_delimiter_or_quotes = False # Paired " is got, so save the second "
        elif flag_expect_delimiter_or_quotes:
            raise Exception(f"Expected separator or second \", but received '{ch}' in offset {i} of '{line}'")
        field_value += ch
    fields_in_the_row.append(strip_column(field_value))
    return fields_in_the_row
# ******************************************************************************
def load_csv(
    file_name: str,
    format:list = None,
    separator:str = ',',
    contains_header:bool = False,
    strip_line:callable = lambda line: line,
    strip_column:callable = lambda column_value: column_value,
    parse_csv_line:callable = parse_complex_csv_line,
    skip_empty_lines:bool = True,
) -> typing.Union[list, dict]:

    # n0debug("strip_line")
    if not strip_line:
        strip_line = lambda line_value: line_value
    elif isinstance(strip_line, bool) and strip_line == True:
        strip_line = lambda line_value: line_value.strip()

    # n0debug("strip_column")
    if not strip_column:
        strip_column = lambda column_value: column_value
    elif isinstance(strip_column, bool) and strip_column == True:
        strip_column = lambda column_value: column_value.strip()

    # loaded_csv = []
    with open(file_name, 'rt') as in_file:
        if contains_header:
            # n0debug("contains_header")
            #file_offset = 0
            #for line in in_file.readlines():
            #    header_line = strip_line(line.rstrip('\n'))
            #    n0debug("header_line")
            #    if header_line:
            #        break
            #    file_offset = in_file.tell()
            #else:
            #    return [] # Empty file or file only with spaces
            while True:
                file_offset = in_file.tell()
                if not (line:=in_file.readline()):
                    return [] # Empty file or file only with spaces
                # n0debug("line")
                header_line = strip_line(line.rstrip('\n'))
                # n0debug("header_line")
                if header_line:
                    break
            # n0debug("header_line")
            # column_names = header_line.rstrip('\n').split(separator)
            column_names = parse_csv_line(header_line, separator, strip_column)
            # n0debug("column_names")
            if not isinstance(contains_header, bool):
                # if contains_header is not boolean, then first line could be header or not
                # to check if the first line is header, check value of first column
                # if it's like expected, then the first line is header
                if column_names[0] != contains_header:
                    in_file.seek(file_offset) # first line is NOT header, re-read first line as data line
                    column_names = None
            if not format:
                format = column_names
            # n0debug("format")

        if format and len(format) != len(set(format)):
            raise Exception(f"Format {format} contains not unique names of columns")


        # TypeError: 'type' object is not subscriptable

        # for line in in_file.readlines():
        while True:
            if not (line:=in_file.readline()):
                # break # EOF
                return None  # EOF
            # n0debug("line")

            stripped_line = strip_line(line.rstrip('\n'))
            # n0debug("stripped_line")
            if skip_empty_lines and not stripped_line:
                continue
            column_values = parse_csv_line(stripped_line, separator, strip_column)
            if format:
                column_values = n0dict(zip(format, column_values))
            # loaded_csv.append(column_values)
            yield column_values
    # return loaded_csv
# ******************************************************************************
def load_simple_csv(
    file_name: str,
    format:list = None,
    separator:str = ',',
    contains_header = False,
    strip_line:callable = lambda line: line,
    strip_column:callable = lambda column_value: column_value,
) -> list:
    return load_csv(
        file_name       = file_name,
        format          = format,
        separator       = separator,
        contains_header = contains_header,
        strip_line      = strip_line,
        strip_column    = strip_column,
        parse_csv_line  = lambda line, separator, strip_column: [strip_column(column_value) for column_value in line.split(separator)],
    )
# ******************************************************************************
def load_complex_csv(
    file_name: str,
    format:list = None,
    separator:str = ',',
    contains_header = False,
    strip_line:callable = lambda line: line,
    strip_column:callable = lambda column_value: column_value,
) -> list:
    return load_csv(
        file_name       = file_name,
        format          = format,
        separator       = separator,
        contains_header = contains_header,
        strip_line      = strip_line,
        strip_column    = strip_column,
        parse_csv_line  = parse_complex_csv_line
    )
# ******************************************************************************
def generate_comlex_csv_row(row:list, separator:str = ',') -> str:
    buff = ""
    for column_i,column_value in enumerate(row):
        if column_value is None:
            column_value = ""
        elif not isinstance(column_value, str):
            column_value = str(column_value)
        if column_i:
            buff += separator
        if separator in column_value:
            if '"' in column_value:
                column_value = column_value.replace('"', '""')
            column_value = '"' + column_value + '"'
        buff += column_value
    buff += '\n'
    return buff
# ******************************************************************************
def generate_csv(root_node:dict, list_xpath:str, mapping_dict:dict = None, save_to:str = None, separator:str = ',', show_header:bool = True) -> list:
    '''
    Samples:
        response_json = n0dict({
            "records" : [
                {
                    "node" : {
                        "field1": "row1_value1",
                        "subnode1": {
                            "field2": "row1_value2",
                        },
                        "subnode2": {
                            "field3": "row1_value3",
                        },
                    }
                },
                {
                    "node" : {
                        "field1": "row2_value1",
                        "subnode1": {
                            "field2": "row2_value2",
                        },
                        "subnode2": {
                            "field3": "row2_value3",
                        },
                    }
                },
            ]
        })
        generated_csv = generate_csv(
                                        response_json,
                                        "//records[*]",
                                        {
                                            "Name of field #1": "node/field1",
                                            "Name of field #2": "node/subnode1/field2",
                                            "Name of field #3": ("node/subnode1/field3", "node/subnode2/field3"),
                                        }
        ) == [
            ["row1_value1", "row1_value2", "row1_value3"],
            ["row2_value1", "row2_value2", "row2_value3"],
        ]
        generate_csv(
                                        response_json,
                                        "//records[*]",
                                        {
                                            "Name of field #1": "node/field1",
                                            "Name of field #2": "node/subnode1/field2",
                                            "Name of field #3": "node/subnode2/field3",
                                        },
                                        "sample.csv"
        )
    '''

    if isinstance(root_node, (list, tuple)):
        list_of_items = root_node
    else:
        list_of_items = root_node.get(list_xpath, tuple())
    # n0debug("list_of_items")

    if save_to:
        out_filehandler = open(save_to, 'wt')

    csv_table = []
    if list_of_items:
        if mapping_dict is None:
            if isinstance(list_of_items[0], list):
                header = [str(i) for i in range(0,len(list_of_items[0]))]
            elif isinstance(list_of_items[0], dict):
                header = list(list_of_items[0].keys())
            mapping_dict = {column_name: column_name for column_name in header}
        else:
            header = list(mapping_dict.keys())

        if save_to and show_header:
            out_filehandler.write(generate_comlex_csv_row(header, separator))

        for found_item in list_of_items:  # found_item == row in case of CSV list or item node in case of XML structure
            # n0debug("found_item")
            if isinstance(found_item, dict):
                found_item = n0struct.n0dict(found_item) # to have ability to use .first(xpath)
            elif isinstance(found_item, list):
                found_item = n0struct.n0list(found_item) # to have ability to use .first(xpath)

            csv_row = []
            for column_name in mapping_dict:
                # n0debug("column_name")
                xpaths = mapping_dict[column_name]
                # n0debug("xpaths")
                if not isinstance(xpaths, (list, tuple)):
                    xpaths = [xpaths]
                for xpath in xpaths:
                    # n0debug("xpath")
                    found_value = found_item.first(xpath, "")
                    if found_value:
                        break
                csv_row.append(found_value)
            csv_table.append(csv_row)

            if save_to:
                out_filehandler.write(generate_comlex_csv_row(csv_row, separator))

    if save_to:
        out_filehandler.close

    return csv_table
# ******************************************************************************
def remove_colums_in_csv(columns_to_remove: list, csv_rows: list = None, csv_schema: dict = None, are_required: bool = False):
    if all([isinstance(column_to_remove,int) for column_to_remove in columns_to_remove]) \
       and len(csv_rows) and isinstance(csv_rows[0], list):
        columns_to_remove.sort(reverse=True) # delete firstly last elements in the list


    for row_index, row in enumerate(csv_rows):
        # Remove columns in structure
        if isinstance(csv_rows[row_index], (dict, list)):
            for column_name in columns_to_remove:
                del csv_rows[row_index][column_name]
        else:
            raise Exception(f"row #{row_index} in csv_rows should dict/list, got {type(csv_rows[row_index])}")

    if isinstance(csv_schema, dict):
        for column_name in columns_to_remove:
            # Update schema: Delete columns
            column_schema = csv_schema.get(f"items/[id={column_name}]")
            # n0debug("column_schema")
            csv_schema.delete(f"items/[id={column_name}]")
            # n0debug("csv_schema")


        csv_schema['maxItems'] = csv_schema.get('maxItems', 0) - (len__columns_to_remove:=len(columns_to_remove))
        if are_required:
            csv_schema['minItems'] = csv_schema.get('minItems', 0) - len__columns_to_remove
            if 'required' in csv_schema:
                for column_to_remove in reversed(columns_to_remove):
                    # n0debug_calc(csv_schema['required'], "csv_schema['required']")
                    for i in range(len(csv_schema['required'])-1, -1, -1):
                        # n0debug("i")
                        if csv_schema['required'][i] == column_to_remove:
                            del csv_schema['required'][i]
                            break
        return csv_rows, csv_schema
    else:
        return csv_rows
# ******************************************************************************
def add_colums_into_csv(additional_columns: list, csv_rows: list = None, csv_schema: dict = None, are_required: bool = False):
    for row_index, row in enumerate(csv_rows):
        for column_name in additional_columns:
            # Add additional columns into structure
            if isinstance(csv_rows[row_index], dict):
                csv_rows[row_index].update({column_name: None})
            elif isinstance(csv_rows[row_index], list):
                csv_rows[row_index].append([None]*len(additional_columns))
            else:
                raise Exception(f"row #{row_index} in csv_rows should dict/list, got {type(csv_rows[row_index])}")

    if isinstance(csv_schema, dict):
        for column_name in additional_columns:
            # Update schema: Add additional columns
            csv_schema['items'].append({'id': column_name})

        csv_schema['maxItems'] = csv_schema.get('maxItems', 0) + (len__additional_columns:=len(additional_columns))
        if are_required:
            csv_schema['minItems'] = csv_schema.get('minItems', 0) + len__additional_columns
            if not 'required' in csv_schema:
                csv_schema['required'] = []
            csv_schema['required'].extend(additional_columns)

        return csv_rows, csv_schema
    else:
        return csv_rows
# ******************************************************************************
def validate_csv_row(row: typing.Union[list, dict], csv_schema: dict,
    external_variables = {},
    interrupt_after_first_fail: bool = False,
):
    '''
    row:
        list: [column_value[0], column_value[1], ... column_value[i]]
        dict: {column_name[0]: column_value[0], column_name[1]: column_value[1], ... column_name[i]: column_value[i]}
    csv_schema: dict
        //items[0..i]/id                     = column_name[i]
        //items[0..i]/mandatory              = true|false
        //items[0..i]/validations[0..j@i][0] = mandatory str, which could be processed with eval("lambda column_value, row:" + validations[0..j@i][0]),
                                               for example: "column_value == __params__['ORG']"
                                               in the lambda function body could be used variables:
                                                    column_value
                                                    row[column_name[0..i]]
                                                    global variables
        //items[0..i]/validations[0..j@i][1] = optional f-str, message in case of validations[0..j@i][0] returned False,
                                               for example: "expected '{ORG}', but got '{value_ORG}'"
                                               in the failed message body could be used variables:
                                                    external_variables.keys()
                                                    value_{column_name[0..i]}
        //minItems                           = optional int, min of columns
        //maxItems                           = optional int, max of columns
        //required[0..k]                     = optional str, mandatory column_name[k]
    external_variables: dict
        List of variables with names external_variables.keys() used in failed validation messages (f-str)
        For example __params__ as {
            'ORG': "123",
            'BIN': "456789"
        } could be used to exchange f-str placeholders {ORG} and {BIN}
    '''
    if isinstance(row, list):
        row = {column_index: column_value for column_index, column_value in enumerate(row)}
    elif not isinstance(row, dict):
        raise Exception(f"Incoming agrument 'row' must be list or dict, but received {type(row)} '{row}'")

    the_whole_row_related_validations = None
    failed_validations = {the_whole_row_related_validations: []}
    columns_count = len(row)
    if (minItems:=csv_schema.get("minItems")):
        if columns_count < minItems:
            failed_validations[the_whole_row_related_validations].append(f"Mimimum {minItems} columns expected, but got {columns_count}")
    if (maxItems:=csv_schema.get("maxItems")):
        if columns_count > maxItems:
            failed_validations[the_whole_row_related_validations].append(f"Maximum {maxItems} columns expected, but got {columns_count}")
    if (required_columns:=csv_schema.get("required")):
        for required_column in required_columns:
            if not required_column in row:
                failed_validations[the_whole_row_related_validations].append(f"Required column '{required_column}' doesn't exist in row")
    if not failed_validations[the_whole_row_related_validations]:
        del failed_validations[the_whole_row_related_validations]
    elif interrupt_after_first_fail:
        return failed_validations

    mapped_values = {**external_variables, **{f'value_{column_name}': row[column_name] for column_name in row}}

    # n0debug("csv_schema")
    for column_name in row:
        column_value = row[column_name]
        # n0debug("column_name")
        column_schema = csv_schema.first(f"items/[id={column_name}]")
        # n0debug("column_schema")

        if column_schema.get('mandatory', False) and not column_value:
            failed_validations.update({column_name: f"mandatory column '{column_name}' is empty"})
            continue

        for validation in column_schema.get('validations', ()):
            # n0debug("validation")
            validation_msg = None
            if isinstance(validation, (list,tuple)):
                validation_lambda = validation[0]
                if len(validation) >= 2:
                    validation_msg = validation[1]
            else:
                validation_lambda = validation

            try:
                lambda_function_for_validation = eval("lambda column_value, row: " + validation_lambda)
            except Exception as ex1:
                n0error(f"{validation_lambda=}")
                n0error(f"{ex1=}")

            try:
                validation_result = lambda_function_for_validation(column_value, row)
            except Exception as ex2:
                set_debug_show_object_type(True)
                n0debug("external_variables")
                n0error(f"{validation_lambda=}")
                n0error(f"{ex2=}")

            if not validation_result:
                if not column_name in failed_validations:
                    failed_validations.update({column_name: []})

                try:
                    failed_validation_message = validation_msg.format(**mapped_values)
                except:
                    failed_validation_message = "not passed validation"
                failed_validations[column_name].append(failed_validation_message)
                if interrupt_after_first_fail:
                    return failed_validations

    return failed_validations
# ******************************************************************************
# ******************************************************************************



import typing
# ******************************************************************************
# ******************************************************************************
def split_pair(
                in_str: str,
                separator: str,
                transform1: callable = lambda x:x,
                transform2: callable = lambda x:x,
                default_element: int = 1
) -> tuple:
    """
    split_pair(in_str: str, separator: str, transform1: callable = lambda x:x, transform2: callable = lambda x:x, default_element: int = 1) -> tuple:

    split string into 2 sub strings in any cases:
        '' by '://'                                     => (None,   None)
        'www.aaa.com' by '://'                          => (None,   'www.aaa.com')
        'https://www.aaa.com' by '://'                  => ('http', 'www.aaa.com')
        'www.aaa.com',default_element = 0 by '/'        => ('www.aaa.com')
        'www.aaa.com/path',default_element = 0 by '/'   => ('www.aaa.com', 'path')
    """
    if not in_str:
        return transform1(None), transform2(None)

    str_parts = in_str.split(separator, 1)
    if len(str_parts) == 1:
        if default_element:
            # second (right) element is default
            return transform1(None), transform2(str_parts[0])
        else:
            # first (left) element is default
            return transform1(str_parts[0]), transform2(None)
    return transform1(str_parts[0]), transform2(str_parts[1])
# ******************************************************************************
def join_triplets(
                    in_list: typing.Union[None, str, tuple, list],
                    level:int = 0
) -> str:
    """
    join_triplets(in_list: typing.Union[None, str, tuple, list], level = 0) -> str:

    join elements with middle sepataror:
        [elem1, separator, elem2]   => elem1, separator, elem2
        [elem1, separator, None]    => elem1
        [None, separator, elem2]    => elem2
    """
    if isinstance(in_list, str) or in_list is None:
        return in_list or ""
    elif not isinstance(in_list, (tuple, list)):
        raise Exception(f"{type(in_list)} '{in_list}' should be None|str|tuple|list")
    elif not len(in_list):
        return ""
    elif len(in_list) > 3:
        raise Exception(f"'{in_list}' should consist not more 3 elements")
    else:
        out_list = []
        for item in in_list:
            out_list.append(join_triplets(item, level+1))

        if len(in_list) == 3:
            if isinstance(in_list[1], str):
                result = out_list[0] +  (out_list[1] if out_list[0] and out_list[2] else "") + out_list[2]
            else:
                if not isinstance(in_list[1], (tuple, list)) or not len(in_list[1]) in (1,2):
                    raise Exception(f"{type(in_list[1])} '{in_list[1]}' is not correct separator")
                if len(in_list[1]) == 2 and not in_list[1][0]:
                    result = out_list[0] +  (out_list[1] if out_list[2] else "") + out_list[2]
                else:
                    result = out_list[0] +  (out_list[1] if out_list[0] else "") + out_list[2]
        else:
            result = "".join(out_list)

        return result
# ******************************************************************************
# ******************************************************************************
